#include <inc/mem.h>
#include <inc/dma.h>
#include <inc/eio.h>
#include <inc/pic.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "DMA"
static int log_level;

struct dma_regs {
  int now[2];
  us base[2];
  ub mode;
  ub page;
  ub dack;
  ub eop;
  DMA_read_handler read_handler;
  DMA_misc_handler misc_handler;
};

#define ADDR 0
#define COUNT 1

static struct dma_cont {
  ub status;
  ub command;
  ub mask;
  ub ff;
  struct dma_regs regs[4];
} dma_controllers[2];

enum {
  CMD_MEMORY_TO_MEMORY = 0x01,
  CMD_FIXED_ADDRESS    = 0x02,
  CMD_BLOCK_CONTROLLER = 0x04,
  CMD_COMPRESSED_TIME  = 0x08,
  CMD_CYCLIC_PRIORITY  = 0x10,
  CMD_EXTENDED_WRITE   = 0x20,
  CMD_LOW_DREQ         = 0x40,
  CMD_LOW_DACK         = 0x80,
  CMD_NOT_SUPPORTED    = CMD_MEMORY_TO_MEMORY | CMD_FIXED_ADDRESS
  | CMD_COMPRESSED_TIME | CMD_CYCLIC_PRIORITY | CMD_EXTENDED_WRITE
  | CMD_LOW_DREQ | CMD_LOW_DACK

};

static void log_chan (const char *cap, int ncont, int ichan)
{
  struct dma_regs *r;

  DEREF (cap);
  r = dma_controllers[ncont].regs + ichan;

  linfo ("%s: ncont %d ichan %d\n"
         "%s: baseaddr %6x, %d\n"
         "%s:  nowaddr %6x, %d\n"
         "%s: baseword %6x, %d\n"
         "%s:  nowword %6x, %d\n",
         cap, ncont, ichan,
         cap, r->base[0], r->base[0],
         cap, r->now[0], r->now[0],
         cap, r->base[1], r->base[1],
         cap, r->now[1], r->now[1]
    );
}

static void write_page (int nport, int size, ui val)
{
  int ichan;
  int ncont;
  static int channels[8] = {-1, 2, 3, 1, -1, -1, -1, 0};

  ncont = nport > 0x87;
  ichan = channels[nport - 0x80 - (ncont << 3)];

  EIO_LOG_WRITE;

  if (-1 == ichan) {
    Fail ("(nport=%#x, size=%d, val=%#x)",
          nport, size, val);
  }

  ldebug ("ncont=%d, ichan=%d, val=%d\n", ncont, ichan, val);
  dma_controllers[ncont].regs[ichan].page = val;
}

static void init_chan (int ncont, int ichan)
{
  struct dma_regs *r;

  r = dma_controllers[ncont].regs + ichan;
  r->now[ADDR] = r->base[0] << ncont;
  r->now[COUNT] = 0;
}

static inline int getff (int ncont)
{
  int ff;

  ff = dma_controllers[ncont].ff;
  dma_controllers[ncont].ff = !ff;
  return ff;
}

static ui read_chan (int nport, int size)
{
  int ff;
  int ncont, ichan, read_count;
  struct dma_regs *r;
  int val;

  DEREF (size);

  ncont = nport > 7;
  ichan = (nport >> (1 + ncont)) & 3;
  read_count = (nport >> ncont) & 1;
  r = dma_controllers[ncont].regs + ichan;

  ff = getff (ncont);

  if (read_count)
    val = (r->base[COUNT] << ncont) - r->now[COUNT];
  else
    val = r->now[ADDR] + r->now[COUNT];

  ldebug ("nport=%#x ncont=%d ichan=%d ff=%d val=%#x\n",
          nport, ncont, ichan, ff, val);
  return (val >> (ncont + (ff << 3))) & 0xff;
}

static void write_chan (int nport, int size, ui val)
{
  int ncont, ichan, write_count;
  int ff;
  struct dma_regs *r;

  EIO_LOG_WRITE;

  ncont = nport > 7;
  ichan = (nport >> (1 + ncont)) & 3;
  write_count = (nport >> ncont) & 1;
  r = dma_controllers[ncont].regs + ichan;

  if (2 == size) {
    r->base[write_count] = val;
    init_chan (ncont, ichan);
  }
  else {
    int shift;

    ff = getff (ncont);
    shift = ff << 3;

    val = (val & 0xff) << shift;
    r->base[write_count] = (r->base[write_count] & (0xff << (8 - shift))) | val;

    if (ff)
      init_chan (ncont, ichan);
  }

  log_chan ("write_chan", ncont, ichan);
  return;
}

static void write_cont (int nport, int size, ui val)
{
  int iport, ichan, ncont;
  char *msg;
  struct dma_cont *d;

  msg = (char *)-1;
  ncont = nport > 0xf;
  ichan = -1;

  EIO_LOG_WRITE;

  d = dma_controllers + ncont;
  if (ncont) {
    iport = ((nport - 0xd0) >> 1) + 8;
  }
  else {
    iport = nport;
  }

  switch (iport) {
    case 8:                     /* command */
      if (val && (val | CMD_NOT_SUPPORTED)) {
        msg = "not supported";
        goto error;
      }
      d->command = val;
      break;

    case 9:
      ichan = val & 3;
      if (val & 4) {
        d->status |= 1 << (ichan + 4);
      }
      else {
        d->status &= ~(1 << (ichan + 4));
      }
      d->status &= ~(1 << ichan);
      break;

    case 0xa:                   /* single mask */
      if (val & 4)
        d->mask |= 1 << (val & 3);
      else
        d->mask &= ~(1 << (val & 3));
      break;

    case 0xb:                   /* mode */
    {
      int op;
      int ai;
      int dir;
      int opmode;

      ichan = val & 3;
      op = (val >> 2) & 3;
      ai = (val >> 4) & 1;
      dir = (val >> 5) & 1;
      opmode = (val >> 6) & 3;

      linfo ("ichan %d, op %d, ai %d, dir %d, opmode %d\n",
             ichan, op, ai, dir, opmode);

#if 0
      if (dir) {
        msg = "decrement";
        goto error;
      }

      if (0 == ai) {
        msg = "non autoinit";
        goto error;
      }

      if (2 != op) {
        msg = "not read";
        goto error;
      }

      if (1 != opmode) {
        msg = "non single";
        goto error;
      }
#endif
      d->regs[ichan].mode = val;
      break;
    }

    case 0xc:                     /* clear flip flop */
      d->ff = 0;
      break;

    case 0xd:                   /* reset */
      d->ff = 0;
      d->mask = ~0;
      d->status = 0;
      d->command = 0;
      break;

    case 0xe:                   /* clear mask for all channels */
      d->mask = 0;
      break;

    case 0xf:                   /* write mask for all channels */
      d->mask = val;
      break;


    default:
      msg = "hello??";
      goto error;
  }

  if (0xc != iport) {
    linfo ("nport %#06x, ncont %d, ichan % 2d, val %#06x\n",
           nport, d != dma_controllers, ichan, val);
  }
  return;

 error:
  Fail ("(nport=%#x, size=%d, val=%u) %s",
        nport, size, val, msg);
}

int DMA_get_channel_mode (int nchan)
{
  return dma_controllers[nchan > 3].regs[nchan & 3].mode;
}

void DMA_hold_DREQ (int nchan)
{
  int ncont, ichan;

  ncont = nchan > 3;
  ichan = nchan & 3;
  linfo ("held cont=%d chan=%d\n", ncont, ichan);
  dma_controllers[ncont].status |= 1 << (ichan + 4);
}

void DMA_release_DREQ (int nchan)
{
  int ncont, ichan;

  ncont = nchan > 3;
  ichan = nchan & 3;
  linfo ("released cont=%d chan=%d\n", ncont, ichan);
  dma_controllers[ncont].status &= ~(1 << (ichan + 4));
}

static void channel_run (int ncont, int ichan)
{
  struct dma_regs *r;
  int n;
  /* int ai, dir; */
  int irq;
  ui addr;

  r = dma_controllers[ncont].regs + ichan;
/*   ai = (r->mode & 16); */
/*   dir = r->mode & 32 ? -1 : 1; */

  addr = MEM_REAL ((r->page << 16) | r->now[ADDR]);

  irq = -1;
  n = r->read_handler (addr, (r->base[COUNT] << ncont) + (1 << ncont), &irq);
  r->now[COUNT] = n;

  ldebug ("dma_pos %d irq %d size %d\n",
          n, irq, (r->base[1] << ncont) + (1 << ncont));

  if (-1 != irq) {
    PIC_request_service (irq);
  }
}

void DMA_run ()
{
  static int in_dma;
  struct dma_cont *d;
  int icont, ichan;

  Fail_if (in_dma);

  in_dma = 1;
  d = dma_controllers;

  for (icont = 0; icont < 2; icont++, d++) {
    for (ichan = 0; ichan < 4; ichan++) {
      int mask;

      mask = 1 << ichan;

      if ((0 == (d->mask & mask)) && (0 != (d->status & (mask << 4))))
        channel_run (icont, ichan);
    }
  }
  in_dma = 0;
}

void DMA_register_channel (int nchan,
                           DMA_read_handler read_handler,
                           DMA_misc_handler misc_handler)
{
  struct dma_regs *r;
  int ichan, ncont;

  ncont = nchan > 3;
  ichan = nchan & 3;

  r = dma_controllers[ncont].regs + ichan;
  r->read_handler = read_handler;
  r->misc_handler = misc_handler;
}

void DMA_init ()
{
  int i;
  int page_port_list[] = { 0x1, 0x2, 0x3, 0x7 };

  log_level = MUP_get_log_level ("dma");

  for (i = 0; i < 8; i++) {
    EIO_register_write_handler (i, EIO_MASK_BW, write_chan);
    EIO_register_write_handler (0xc0 + (i << 1), EIO_MASK_BW, write_chan);

    EIO_register_read_handler (i, EIO_MASK_B, read_chan);
    EIO_register_read_handler (0xc0 + (i << 1), EIO_MASK_B, read_chan);
  }

  for (i = 0; i < LENOFA (page_port_list); i++) {
    EIO_register_write_handler (page_port_list[i] + 0x80, EIO_MASK_B, write_page);
    EIO_register_write_handler (page_port_list[i] + 0x88, EIO_MASK_B, write_page);
  }

  for (i = 0; i < 8; i++) {
    EIO_register_write_handler (i + 8, EIO_MASK_B, write_cont);
    EIO_register_write_handler (0xd0 + (i << 1), EIO_MASK_B, write_cont);
  }

  write_cont (0xd, 1, 0);
  write_cont (0xdd, 1, 0);
}
