#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/stat.h>

#include <inc/mem.h>
#include <inc/ems.h>
#include <inc/dos.h>
#include <inc/vga.h>
#include <inc/kbd.h>
#include <inc/int.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "DOS"
static int log_level;

#define LOL_SIZE                           76
#define LOL_FIRST_MCB(lol)                 (*(us*)((lol)-2))

#define MCB_ID(mcb)                        (*(ub*)(mcb))
#define MCB_PSP_OWNER_SEG(mcb)             (*(us*)(mcb+1))
#define MCB_SIZE(mcb)                      (*(us*)(mcb+3))
#define MCB_FREE(mcb)                      (*(ub*)(mcb+5))
#define MCB_OF_LIN(linear)                 ((ub *)(linear-16))

#define PSP_EXIT(psp)                      (*(us*)(psp))
#define PSP_NEXT_SEGMENT(psp)              (*(us*)(psp+2))
#define PSP_CPM_ENTRY_POINT(psp)           ((ub*)(psp+5))
#define PSP_FIRST_SEG_SIZE_FOR_COM(psp)    (*(us*)(psp+6))
#define PSP_INT22(psp)                     (*(ui*)(psp+0xa))
#define PSP_INT23(psp)                     (*(ui*)(psp+0xe))
#define PSP_INT24(psp)                     (*(ui*)(psp+0x12))
#define PSP_PARENT_PSP(psp)                (*(ui*)(psp+0x16))
#define PSP_JFT(psp)                       ((ub*)(psp+0x18))
#define PSP_SSSP_ON_LAST_INT21CALL(psp)    (*(ui*)(psp+0x2e))
#define PSP_NUMBER_OF_ENTRIES_IN_JFT(psp)  (*(us*)(psp+0x32))
#define PSP_JFT_POINTER(psp)               (*(ui*)(psp+0x34))
#define PSP_PREVIOUS_PSP(psp)              (*(ui*)(psp+0x38))
#define PSP_COMMAND_LINE(psp)              ((ub*)(psp+0x80))

#define MCB_LAST_BLOCK                     0x5a
#define MCB_NOT_LAST_BLOCK                 0x4d
#define LOL_START MEM_REAL (0x1000)
#define LOL (LOL_START + 24)
/* #define ROOT_MCB                           ((ub*)ROOT_MCB_LIN) */
#define ROOT_MCB                           ((ub *) MEM_REAL (0x5000))

#define DTA_DP(dta)                        (*(DIR**)(dta))
#define DTA_DIR(dta)                       (*(char**)(dta+4))
#define DTA_MSK(dta)                       (*(char**)(dta+8))
#define DTA_ATR(dta)                       (*(us *)(dta+12))
#define DTA_FILE_ATTR(dta)                 (*(ub*)(dta+0x15))
#define DTA_FILE_TIME(dta)                 (*(us*)(dta+0x16))
#define DTA_FILE_DATE(dta)                 (*(us*)(dta+0x18))
#define DTA_FILE_SIZE(dta)                 (*(ui*)(dta+0x1a))
#define DTA_FILE_NAME(dta)                 ((char*)(dta+0x1e))

#define NEXT_MCB(mcb)                      (mcb+((1+MCB_SIZE(mcb))<<4))

static int con_fd = -1;
static int env_size;
static us PSP_seg;
static us env_seg;
static char *dta;
static int dta_dirty;
static struct file_info {
  int fd;
  char *path;
  us diw;
  off_t pos;
  off_t size;
} file_handles[20];

#define MAXDRIVES ('Z' - 'A' + 1)
static struct fs_drive {
  char *path;
  char *cwd;
} fs_drives[MAXDRIVES] = {{0, 0}, };

static int fs_ndrive;
static int fs_maped_drives;
static int fs_highest_maped_drive;
static doserr_e dos_errno;

static char **env_strings;
static int nenv_strings;
static int nenv_strings_allocated;

enum {
  HCON = 2,
  HNUL = 3,
  HPRN = 4,
  HAUX = 5,
  HCIN = 6,
  HEMM = 7,
};

enum {
  FILE_IS_REMOTE = 1 << 15,
  FILE_DO_NOT_SET_FILE_DATE_TIME_ON_CLOSING = 1 << 14,
  FILE_MEDIA_NOT_REMOVABLE = 1 << 11,
  FILE_HAS_NOT_BEEN_WRITTEN = 1 << 6,
};

static void set_error (doserr_e code)
{
  dos_errno = code;
}

doserr_e DOS_get_error ()
{
  return dos_errno;
}

static char *strrpbrk (const char *s, const char *accept)
{
  char *res;
  char *w;

  res = NULL;

  w = strpbrk (s, accept);
  while (w > res) {
    res = w;
    w = strpbrk (w + 1, accept);
  }
  return res;
}

/*
 * FS utility code
 */
static struct fs_drive *get_fs_drive (int ndrive)
{
  if ((ndrive < 0) || (ndrive >= MAXDRIVES)) {
    Aldri ();
  }
  return fs_drives + ndrive;
}

static void log_fs (const char *cap)
{
  struct fs_drive *fs_drive;

  DEREF (cap);
  fs_drive = get_fs_drive (fs_ndrive);
  ldebug ("%s\n"
          "fs_drive %d\n"
          "path `%s' %p\n"
          "cwd  `%s' %p\n",
          cap,
          fs_ndrive,
          fs_drive->path, fs_drive->path,
          fs_drive->cwd, fs_drive->cwd);
}

#define FILE_toupper toupper
#include <inc/wfm.inc>

int DOS_get_ndrives ()
{
  return fs_maped_drives;
}

int DOS_get_highest_drive ()
{
  return fs_highest_maped_drive;
}

int DOS_get_drive ()
{
  return fs_ndrive;
}

int DOS_set_drive (int drive)
{
  ldebug ("set drive %d=%c old %d=%c\n",
          drive, drive + 'A', fs_ndrive, fs_ndrive + 'A');
  fs_ndrive = drive;
  return 0;
}

int DOS_get_fs_map_info (int drive, char **map_path, char **wd)
{
  struct fs_drive *fs_drive;

  fs_drive = get_fs_drive (drive);

  if (fs_drive->path) {
    *map_path = fs_drive->path;
    *wd = fs_drive->cwd;
    return 0;
  }
  else {
    return -1;
  }
}

/*
 * FS code
 */
static void replace_slashes (char *s)
{
  while ((s = strchr (s, '/'))) {
    *s++ = '\\';
  }
}

static void setup_drive (int drive, char *unix_dir)
{
  struct fs_drive *fs;

  fs_ndrive = drive;
  fs = get_fs_drive (drive);
  maybe_free (fs->cwd);
  fs->cwd = unix_dir;
}

static doserr_e transl (const char *dos_path,
                        char **unix_path,
                        int *dos_drive)
{
  int was_blocked;
  int drive;
  size_t dos_path_len;
  int starting_from_root;
  size_t result_len;
  int root_len;
  doserr_e retval;
  size_t current_len;
  char *result;
  const char *s;
  struct fs_drive *fs;

# define starting_from_current (0 == starting_from_root)
# define Is_Slash(c) ('\\' == (c) || '/' == (c))

  *unix_path = NULL;
  *dos_drive = -1;

  s = NULL;
  fs = NULL;
  result = NULL;

  ldebug ("(dos_path=`%s')\n", dos_path);

  dos_path_len = strlen (dos_path);

  if ((dos_path_len > 1) && (':' == dos_path[1])) {
    drive = toupper (dos_path[0]) - 'A';

    if (dos_path_len > 2 && Is_Slash (dos_path[2])) {
      starting_from_root = 1;
      s = dos_path + 3;
    }
    else {
      starting_from_root = 0;
      s = dos_path + 2;
    }
  }
  else {
    drive = fs_ndrive;

    if (Is_Slash (dos_path[0])) {
      s = dos_path + 1;
      starting_from_root = 1;
    }
    else {
      s = dos_path;
      starting_from_root = 0;
    }
  }

  fs = get_fs_drive (drive);

  if (NULL == fs->path)
    goto drive_not_maped;

  root_len = strlen (fs->path);
  current_len = strlen (fs->cwd);

  /* calculate maximal resulting path length */

  if ((0 != starting_from_root) || (0 == current_len)) {
    result_len = dos_path_len + root_len + 1;
  }
  else {
    result_len = dos_path_len + current_len + 1;
  }

  /* allocate and fill result */
  result = xmalloc (result_len + 1);

  memset (result, 0, result_len + 1);
  if (starting_from_root) {
    strcpy (result, fs->path);
  }
  else {
    strcpy (result, fs->cwd);
  }

  if (0 == *s)
    goto just_root;

  do {
    DIR *dir;
    struct dirent *dirent;
    unsigned int range_len;
    char *slash_pos;

    while (*s && Is_Slash (*s))
      s++;

    slash_pos = strpbrk (s, "\\/");

    if (slash_pos) {
      range_len = (slash_pos - s);
    }
    else {
      range_len = dos_path_len - (s - dos_path);
    }

    if ((0 != range_len) && ('.' == s[0])) {
      switch (range_len) {
        case 1:
          goto next_iteration;

        case 2: {
          if ('.' == s[1]) {
            char *sp;

            sp = strrpbrk (result, "\\/");
            if (NULL == sp) {
              Fail ("up level `%s'", dos_path);
            }

            if (sp - result < root_len)
              goto higher_than_root;

            *sp = 0;
            if ((NULL == slash_pos) && ((int) (sp - result) == root_len))
              goto just_root;
            else
              goto next_iteration;
          }
        }
      }
    }

    was_blocked = SIG_block ();
    dir = opendir (result);     /* glibc uses calloc internally */
    SIG_unblock (was_blocked);

    if (NULL == dir) {
      ERRFail ("opendir `%s'", result);
    }

    while ((dirent = readdir (dir))) { /* XXX */
      if ((0 == dirent->d_name[range_len])
          && (0 == strncasecmp (dirent->d_name, s, range_len))) {

        strcat (result, "/");
        strcat (result, dirent->d_name);
        closedir (dir);
        goto next_iteration;
      }
    }

    if (slash_pos) {
      goto path_not_found;
    }
    else {
      strcat (result, "/");
      strcat (result, s);
      retval = FILE_NOT_FOUND;
      break;
    }

  next_iteration:
    if (NULL == slash_pos) {
      retval = NO_ERROR;
      break;
    }

    s += range_len + 1;
  } while (1);

  *unix_path = result;
  *dos_drive = drive;
  return retval;

 higher_than_root:
  lwarn ("path `%s' leads outside of map\n", dos_path);

 just_root:
  *unix_path = xstrdup (get_fs_drive (drive)->path);
  *dos_drive = drive;
  xfree (result);
  lwarn ("root `%s'\n", *unix_path);
  return NO_ERROR;

 path_not_found:
  lwarn ("path `%s' not found\n", dos_path);
  retval = PATH_NOT_FOUND;
  goto error;

 drive_not_maped:
  lwarn ("drive %c not maped\n", drive + 'A');
  retval = INVALID_DRIVE;
  goto error;

 error:
  maybe_free (result);
  *unix_path = NULL;
  *dos_drive = -1;
  return retval;
# undef Is_Slash
# undef starting_from_current
}

static void maybe_free_dta ()
{
  if (0 && dta_dirty) {
    DIR *ff_dp;
    char *ff_mask;
    char *ff_dir;

    ff_dp = DTA_DP (dta);
    ff_dir = DTA_DIR (dta);
    ff_mask = DTA_MSK (dta);

    closedir (ff_dp);
    xfree (ff_dir);
    xfree (ff_mask);
    dta_dirty = 0;
  }
}

void DOS_set_DTA (void *new_dta)
{
  if ((char *) new_dta == dta)
    return;

  maybe_free_dta ();
  dta = (char *) new_dta;
}

void *DOS_get_DTA ()
{
  return dta;
}

int DOS_find_next ()
{
  int retval;
  struct dirent *de;
  DIR *ff_dp;
  char *ff_mask;
  char *ff_dir;
  char *stat_str;
  int ff_attr;
  size_t dir_len;

  if (0 == dta_dirty) {
    Fail ("dta not dirty");
  }

  ff_dp = DTA_DP (dta);
  ff_dir = DTA_DIR (dta);
  ff_mask = DTA_MSK (dta);
  ff_attr = DTA_ATR (dta);

  dir_len = strlen (ff_dir);
  stat_str = xmalloc (dir_len + 14);
  memset (stat_str, 0, dir_len);
  strcpy (stat_str, ff_dir);
  stat_str[dir_len] = '/';

  while ((de = readdir (ff_dp))) { /* XXX */
    int entry_len;
    struct stat buf;
    ub attr;
    int sysret;

    entry_len = strlen (de->d_name);

    if (entry_len > 12)
      continue;

    strcpy (stat_str + dir_len + 1, de->d_name);

    sysret = lstat (stat_str, &buf);

    if (-1 == sysret) {
      ERRFail ("lstat `%s'", stat_str);
    }

    if (S_ISDIR (buf.st_mode))
      attr = DIRECTORY;
    else
      attr = ARCHIVE;

    if (((0 == ff_attr) || (attr & ff_attr))
        && DOSFS_MatchLong (ff_mask, de->d_name, 0)) {

      DTA_FILE_ATTR (dta) = attr;
      DTA_FILE_TIME (dta) = 0;
      DTA_FILE_DATE (dta) = 0;
      DTA_FILE_SIZE (dta) = buf.st_size;

      memset (DTA_FILE_NAME (dta), 0, 13);
      strcpy (DTA_FILE_NAME (dta), de->d_name);

      linfo ("found `%s'\n", de->d_name);
      retval = 0;
      goto ok;
    }
  }

  set_error (NO_MORE_FILES);
  maybe_free_dta ();
  retval = -1;

 ok:
  xfree (stat_str);
  return retval;
}

int DOS_find_first (const char *mask, int attr)
{
  char *dir;
  char *unix_path;
  char *s;
  char *m;
  int drive;
  int was_blocked;

  maybe_free_dta ();            /* XXX */

  ldebug ("(mask=`%s',attr=%d)\n", mask, attr);

  m = xstrdup (mask);
  s = strrpbrk (m, "\\/");

  dir = (char *) -1;

  if (s) {
    doserr_e doserr;

    DTA_MSK (dta) = xstrdup (s + 1);
    s[0] = 0;

    doserr = transl (m, &unix_path, &drive);
    xfree (m);

    switch (doserr) {
      case NO_ERROR:
        break;

      default:
        set_error (doserr);
        xfree (DTA_MSK (dta));
        maybe_free (unix_path);
        return -1;
    }
    dir = unix_path;
  }
  else {
    DTA_MSK (dta) = m;
    drive = fs_ndrive;
    dir = xstrdup (get_fs_drive (fs_ndrive)->cwd);
  }

  DTA_ATR (dta) = attr;
  DTA_DIR (dta) = dir;

  was_blocked = SIG_block ();
  DTA_DP (dta) = opendir (dir); /* glibc uses calloc internally */
  SIG_unblock (was_blocked);

  if (NULL == DTA_DP (dta)) {
    ERRFail ("opendir `%s'", dir);
  }

  dta_dirty = 1;
  return DOS_find_next ();
}

int DOS_make_dir (char *dos_path)
{
  doserr_e doserr;
  int ret;
  int drive;
  char *unix_path;

  ldebug ("(dos_path=`%s')\n", dos_path);

  doserr = transl (dos_path, &unix_path, &drive);

  switch (doserr) {
    case NO_ERROR:
      return 0;

    case PATH_NOT_FOUND:
      xfree (unix_path);
      set_error (PATH_NOT_FOUND);
      return -1;

    case FILE_NOT_FOUND:
      break;

    default:
      set_error (doserr);
      return -1;
  }

#ifdef _WIN32
  ret = mkdir (unix_path);
#else
  ret = mkdir (unix_path, 0777);
#endif

  if (-1 == ret) {
    int err;

    err = errno;
    lwarn ("(dos_path=`%s')\n"
           "unix_path=`%s'\n"
           "syserr %s\n",
           dos_path, unix_path, strerror (err));
    set_error (ACCESS_DENIED);
    return -1;
  }

  return 0;
}

int DOS_set_cur_dir (char *dos_path)
{
  doserr_e doserr;
  int drive;
  char *unix_path;

  ldebug ("(dos_path=`%s')\n", dos_path);

  doserr = transl (dos_path, &unix_path, &drive);

  switch (doserr) {
    case NO_ERROR:
      break;

    case PATH_NOT_FOUND:
      xfree (unix_path);
      set_error (PATH_NOT_FOUND);
      return -1;

    default:
      set_error (doserr);
      return -1;
  }

  setup_drive (drive, unix_path);
  return 0;
}

int DOS_remove_dir (char *dos_path)
{
  doserr_e doserr;
  int drive;
  char *unix_path;

  ldebug ("(dos_path=`%s')\n", dos_path);

  doserr = transl (dos_path, &unix_path, &drive);

  switch (doserr) {
    case NO_ERROR:
      break;

    case PATH_NOT_FOUND:
      xfree (unix_path);
      set_error (PATH_NOT_FOUND);
      return -1;

    default:
      set_error (doserr);
      return -1;
  }

  if (-1 == rmdir (unix_path)) {
    int err;

    err = errno;
    lwarn ("rmdir(`%s') unix_path=`%s' failed\n"
           "syserr %s\n",
           dos_path, unix_path,
           strerror (err));
    return -1;
  }
  return 0;
}

int DOS_get_cur_dir (int drive_code, char *buf)
{
  int unix_len;
  char *unix_path;
  struct fs_drive *fs;
  int drive;

  ldebug ("(drive_code=%d)\n", drive_code);

  drive = (0 == drive_code) ? fs_ndrive : drive_code - 1;

  fs = get_fs_drive (drive);

  if (NULL == fs->path) {
    Fail ("drive %c is not maped",
          drive + 'A');
  }

  memset (buf, 0, 64);
  log_fs (__func__);

  unix_path = fs->cwd + strlen (fs->path);
  unix_len = strlen (unix_path);

  if (unix_len > 63) {
    Fail ("unix_path `%s' too big %d max 63", unix_path, unix_len);
  }

  switch (*unix_path) {
    case '\0':
      return 1;

    case '/':
      unix_path++;

    default:
      strcpy (buf, unix_path);
      replace_slashes (buf);
  }

  return 0;
}

void DOS_map_default_drive (char *path)
{
  int i;
  int avail;
  char *dir;
  struct fs_drive *fs_drive;

  avail = -1;

  dir = NULL;

  for (i = 0; i < MAXDRIVES; i++) {
    fs_drive = fs_drives + i;

    if (fs_drive->path) {
      size_t len;

      len = strlen (fs_drive->path);

      if (0 == strncmp (fs_drive->path, path, len)) {
        size_t path_len;
        size_t last_slash_compensate;

        path_len = strlen (path);
        last_slash_compensate = '/' != fs_drive->path[len - 1];

        if (len + last_slash_compensate >= path_len) {
          dir = NULL;
        }
        else {
          dir = path + len + last_slash_compensate;
        }
        avail = i;
        break;
      }
    }
    else {
      if ((-1 == avail) || (avail < 2))
        avail = i;
    }
  }

  if (-1 == avail) {
    Fail ("can not establish mapping for `%s'", path);
  }

  fs_ndrive = avail;
  fs_drive = get_fs_drive (fs_ndrive);

  if (NULL == fs_drive->path) {
    fs_drive->path = xstrdup (path);
    fs_maped_drives++;
    fs_highest_maped_drive = avail;
  }

  if (NULL == dir) {
    dir = xstrdup (fs_drive->path);
  }
  else {
    char *tmp;
    size_t root_len, dir_len;

    root_len = strlen (fs_drive->path);
    dir_len = strlen (dir);
    tmp = xmalloc (root_len + dir_len + 2);
    strcpy (tmp, fs_drive->path);
    strcat (tmp, "/");
    strcat (tmp, dir);
    dir = tmp;
  }
  setup_drive (fs_ndrive, dir);

  linfo ("mapping to %c:\nmap=%s\nwd=%s\n",
         avail + 'A', fs_drive->path, fs_drive->cwd);
}

/*
 * Memory utility
 */
static void log_block (ub *mcb)
{
  DEREF (mcb);
  log ("AT %p NEXT %p\n"
       "ID %c\n"
       "SIZE %#x, %d\n"
       "FREE %d\n\n",
       mcb, 0x10 + mcb + (MCB_SIZE (mcb) << 4),
       MCB_ID (mcb),
       MCB_SIZE (mcb) << 4,
       MCB_SIZE (mcb) << 4,
       MCB_FREE (mcb));
}

static void log_blocks (const char *cap)
{
  if (LOG_DEBUG == log_level) {
    ub *mcb;

    log ("%s\n", cap);
    mcb = ROOT_MCB;

    for (; ; mcb = NEXT_MCB (mcb))
    {
      log_block (mcb);
      if (MCB_LAST_BLOCK == MCB_ID (mcb))
        break;
    }
  }
}

static int coalesce ()
{
  ub *free_mcb, *mcb;
  us size;

  mcb = ROOT_MCB;

  if (MCB_LAST_BLOCK == MCB_ID (mcb))
    return 0;

  free_mcb = NULL;
  size = 0;

  log_blocks ("coalesce");
  while (MCB_LAST_BLOCK != MCB_ID (mcb)) {
    ub *next_mcb;

    if (MCB_NOT_LAST_BLOCK != MCB_ID (mcb)) {
      lwarn ("memory control blocks destroyed\n");
      return -1;
    }

    next_mcb = NEXT_MCB (mcb);

    if (MCB_FREE (mcb)) {
      /* case 1: last block was also free
         increment size

         case 2: last block wasnt free
         this is the first in a row
         set free_mcb
         clear size */

      if (free_mcb) {
        size += MCB_SIZE (mcb) + 1;
      }
      else {
        free_mcb = mcb;
        size = 0;
      }
    }
    else {
      /* case 1: last wasnt free nop
         case 2: last was free, fix it and continue */

      if (free_mcb) {
        MCB_SIZE (free_mcb) += size;
        free_mcb = NULL;
      }
    }

    mcb = next_mcb;
  }

  if (free_mcb) {
    if (MCB_FREE (mcb)) {
      MCB_ID (free_mcb) = MCB_LAST_BLOCK;
      size += MCB_SIZE (mcb) + 1;
    }
    MCB_SIZE (free_mcb) += size;
  }

  return 0;
}

/*
 * Memory
 */
us DOS_largest_block ()
{
  ub *mcb;
  us largest;

  largest = 0;

  for (mcb = ROOT_MCB; ; mcb = NEXT_MCB (mcb))
  {
    if (MCB_FREE (mcb) && MCB_SIZE (mcb) > largest)
      largest = MCB_SIZE (mcb);

    if (MCB_LAST_BLOCK == MCB_ID (mcb))
      break;

    if (MCB_NOT_LAST_BLOCK != MCB_ID (mcb)) {
      return 0;
    }
  }

  return largest;
}

int DOS_alloc (us nparas, ui *linear)
{
  ub *mcb;

  if (-1 == coalesce ()) {
    set_error (MEMORY_CONTROL_BLOCK_DESTROYED);
    return -1;
  }

  if (nparas > DOS_largest_block ()) {
    lwarn ("attempt to get %d paras of DOS memory max block %d\n",
           nparas, DOS_largest_block ());
    set_error (INSUFFICIENT_MEMORY);
    return -1;
  }

  for (mcb = ROOT_MCB; ; mcb = NEXT_MCB (mcb)) {

    if (MCB_FREE (mcb) && MCB_SIZE (mcb) >= nparas) {
      us remaining;
      ub *new_mcb;

      remaining = MCB_SIZE (mcb) - nparas;
      MCB_FREE (mcb) = 0;

      if (remaining > 1) {
        new_mcb = mcb + ((nparas + 1) << 4);
        MCB_FREE (new_mcb) = 1;
        MCB_SIZE (new_mcb) = remaining - 1;
        MCB_ID (new_mcb) = MCB_ID (mcb);
        MCB_PSP_OWNER_SEG (new_mcb) = PSP_seg;

        MCB_SIZE (mcb) = nparas;
        MCB_ID (mcb) = MCB_NOT_LAST_BLOCK;
      }

      *linear = 16 + (ui) mcb;
      linfo ("(nparas=%#x) = %#x\n", nparas, *linear);
      log_blocks ("alloc");
      return 0;
    }

    if (MCB_LAST_BLOCK == MCB_ID (mcb)) {
      set_error (MEMORY_CONTROL_BLOCK_DESTROYED);
      return -1;
    }
  }
}

int DOS_resize (ui linear, us nparas)
{
  ub *mcb;

  ldebug ("(linear=%#x, nparas=%#x)\n",
          linear, nparas);

  log_blocks ("resize");
  if (-1 == coalesce ()) {
    set_error (MEMORY_CONTROL_BLOCK_DESTROYED);
    return -1;
  }

  mcb = MCB_OF_LIN (linear);
  ldebug ("psp=%#x old=%#x new=%#x max=%#x\n",
          PSP_seg, MCB_SIZE (mcb), nparas, DOS_largest_block ());

  if (nparas == MCB_SIZE (mcb)) {
    ldebug ("mcb size == nparas\n");
    return 0;
  }

  if (MCB_SIZE (mcb) > nparas) {
    us remaining;
    ub *new_mcb;

    remaining = MCB_SIZE (mcb) - nparas;
    if (remaining > 1) {
      new_mcb = mcb + ((nparas + 1) << 4);

      MCB_SIZE (mcb) = nparas;

      MCB_ID (new_mcb) = MCB_ID (mcb);
      MCB_SIZE (new_mcb) = remaining - 1;
      MCB_FREE (new_mcb) = 1;
      MCB_PSP_OWNER_SEG (new_mcb) = PSP_seg;
      MCB_ID (mcb) = MCB_NOT_LAST_BLOCK;
    }
    log_blocks ("resize shrink");
    ldebug ("shrink ok\n");
    return 0;
  }
  else {
    us next_paras;
    ub *next_mcb;

    next_mcb = NEXT_MCB (mcb);
    next_paras = MCB_SIZE (next_mcb);

    if (MCB_FREE (next_mcb) && (next_paras + 1 + MCB_SIZE (mcb) >= nparas)) {
      us full, remaining;

      full = MCB_SIZE (mcb) + next_paras + 1;
      remaining = full - nparas;

      /* case 1: we have a space to create one more mcb
         nparas - both blocks size + 1 >= 2
         set mcb's size to nparas
         create new mcb

         case 2: no space
         set mcb's size to both blocks size + 1
      */

      /* BPT; */
      if (remaining > 1) {
        ub *new_mcb;

        new_mcb = mcb + ((nparas + 1) << 4);
        MCB_ID (new_mcb) = MCB_ID (next_mcb);
        MCB_SIZE (new_mcb) = remaining - 1;
        MCB_FREE (new_mcb) = 1;
        MCB_PSP_OWNER_SEG (new_mcb) = PSP_seg;

        MCB_ID (mcb) = MCB_NOT_LAST_BLOCK;
        MCB_SIZE (mcb) = nparas;
      }
      else {
        MCB_ID (mcb) = MCB_ID (next_mcb);
        MCB_SIZE (mcb) = full;
      }

      log_blocks ("resize grow");
      return 0;
    }
    else {
      set_error (INSUFFICIENT_MEMORY);
      return -1;
    }
  }
}

void DOS_free (ui linear)
{
  ub *mcb;

  ldebug ("(linear=%#x)\n", linear);
  mcb = MCB_OF_LIN (linear);
  MCB_FREE (mcb) = 1;
  log_blocks ("resize free");
}

void DOS_put_name (char *name)
{
  char *env;
  char *cwd;
  struct fs_drive *fs_drive;
  char *s;

  env = (char *) MEM_REAL (env_seg << 4);
  s = (char *) env;
  fs_drive = get_fs_drive (fs_ndrive);
  cwd = fs_drive->cwd + strlen (fs_drive->path);

  do {
    s = strchr (s, 0);
  } while (*++s);
  s++;
  *(us *) s = 1;
  s += 2;
  *s++ = fs_ndrive + 'A';
  *s++ = ':';

  if (*cwd) {
    strcpy (s, cwd);
    replace_slashes (s);
  }
  strcat (s, "\\");
  strcat (s, name);
}

void DOS_put_cmdline (char *cmdline)
{
  int len;
  char *psp;

  psp = (char *) MEM_REAL (PSP_seg << 4);
  len = strlen (cmdline);
  if (len > 0x79) {
    Fail ("command line to long(%d) max 127", len);
  }
  psp[0x80] = len;
  strcpy (psp + 0x81, cmdline);
  psp[0x81 + len] = 0xd;
}

ui DOS_get_PSP ()
{
  return MEM_REAL (PSP_seg << 4);
}

void DOS_create_PSP (us seg)
{
  char *lin;

  lin =  (char *) MEM_REAL (seg << 4);
  memset (lin, 0, 0x100);

  PSP_EXIT (lin) = 0x20cd;
  PSP_PARENT_PSP (lin) = 0;
  PSP_PREVIOUS_PSP (lin) = 0xffffU;
  PSP_ENVIRONMENT_SEGMENT (lin) = env_seg;
}

void DOS_set_PSP (us seg)
{
  PSP_seg = seg;
}

#if 0
struct epb {
  us env_seg;
  us cmd_tail_off, cmd_tail_seg;
  us fcb1_off, fcb1_seg;
  us fcb2_off, fcb2_seg;
  us sp, ss;
  us cs, ip;
};

struct opb {
  us ovr_seg;
  us fix;
};
#endif

void DOS_put_env (char *var)
{
  char *env;
  size_t var_len;
  int env_len;
  size_t cmd_len;
  int need;
  char *s;
  char *cmd;

  var_len = strlen (var);

  env = (char *) MEM_REAL (env_seg << 4);
  s = env;
  cmd = NULL;
  cmd_len = 0;
  env_len = 0;

  if (*s) {
    do {
      s = strchr (s, 0);
    } while (*++s);

    cmd = s + 3;
    cmd_len = strlen (cmd);
  }

  if (s != env) {
    env_len = cmd - env + cmd_len + 1;
  }

  need = env_len + var_len + 1;

  if (need > env_size) {
    Fail ("environment to short need %d bytes but got only %d",
          need, env_size);
  }

  if (cmd_len) {
    char *dst;

    dst = cmd + var_len + 1;
    memmove (dst, cmd, cmd_len);
    ((ui *) dst)[-1] = 0;
  }

  memcpy (s, var, var_len + 1);
}

void DOS_con_out (char c)
{
  VGA_print_string_TTY (&c, 1);
}

int DOS_con_in ()
{
  us key;
  static int cache;
  intregs r;

  if (cache) {
    int sc;

    sc = cache;
    cache = 0;
    return sc;
  }

  EAX (r) = 0x1000;
  I16 (r);
  key = EAX (r);

  switch (key & 0xff) {
    case 0: {
      cache = (key >> 8) & 0xff;
      return 0;
    }

    default:
      return key & 0xff;
  }
}

static int con_read (void *buf, int size)
{
  int nread;
  char *p;

  p = buf;
  nread = 0;
  ldebug ("(buf=%p, size=%d)\n", buf, size);

  while (nread < size) {
    int ch;

    ch = DOS_con_in ();
    p[nread] = ch;
    VGA_print_string_TTY (p + nread, 1);
    nread++;

    if (13 == ch) {
      if (nread < size) {
        p[nread++] = 10;
      }
      break;
    }
  }
  return nread;
}

/*
 * IO utility
 */
static int get_free_handle_index ()
{
  unsigned int i;

  for (i = 0; i < LENOFA (file_handles); i++) {
    if (-1 == file_handles[i].fd)
      return i;
  }
  set_error (TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE);
  return -1;
}

static int case_open (const char *dos_path, int mode, int hindex)
{
  int fd;
  int err;
  doserr_e doserr;
  int drive;
  char *unix_path;
  struct file_info *info;

  info = file_handles + hindex;
  ldebug ("(dos_path=`%s', mode=%d)\n", dos_path, mode);

  doserr = transl (dos_path, &unix_path, &drive);

  if (NO_ERROR != doserr) {
    set_error (doserr);
    maybe_free (unix_path);
    return -1;
  }

  ldebug ("unix_path=`%s'\n", unix_path);

  fd = open (unix_path, mode);
  err = errno;
  xfree (unix_path);

  if (-1 == fd) {
    lwarn ("(dos_path=`%s', mode=%d)\n"
           "unix_path=`%s'\n"
           "syserr %s\n",
           dos_path, mode, unix_path, strerror (err));
    set_error (ACCESS_DENIED);
    return -1;
  }

  ldebug ("opened `%s' fd = %d\n", dos_path, fd);

  info->fd = fd;
  info->path = xstrdup (dos_path);
  info->diw = drive | (1 << 6) | (1 << 8) | (1 << 11) | (1 << 14);
  info->pos = 0;
  info->size = lseek (fd, 0, 2);
  if (-1 == info->size) {
    err = errno;
    lwarn ("failed to obtain size of `%s'\n"
           "syserr %s\n", dos_path, strerror (err));
    DOS_close (hindex);
    return -1;
  }

  if (-1 == lseek (fd, 0, 0)) {
    err = errno;
    lwarn ("failed to rewind `%s'\n"
           "syserr %s\n",
           dos_path, strerror (err));
    DOS_close (hindex);
    return -1;
  }
  return fd;
}

static int case_create (const char *dos_path, int attr, struct file_info *info)
{
  int fd;
  doserr_e doserr;
  int err;
  int mode;
  int drive;
  char *unix_path;

  ldebug ("(dos_path=`%s', attr=%d)\n",
          dos_path, attr);

  doserr = transl (dos_path, &unix_path, &drive);

  switch (doserr) {
    case NO_ERROR:
    case FILE_NOT_FOUND:
      break;

    default:
      set_error (doserr);
      return -1;
  }

  if (attr & READ_ONLY)
    mode = O_RDONLY;
  else
    mode = O_RDWR;

  ldebug ("unix_path=`%s'\n", unix_path);

  fd = open (unix_path, O_CREAT | O_TRUNC | mode, 0666);
  err = errno;
  xfree (unix_path);

  if (-1 == fd) {
    lwarn ("(dos_path=`%s', attr=%d)\n"
           "unix_path=`%s'\n"
           "syserr %s\n",
           dos_path, attr, unix_path, strerror (err));
    set_error (ACCESS_DENIED);
    return -1;
  }

  ldebug ("created `%s' fd = %d\n",
          dos_path, fd);

  info->fd = fd;
  info->path = xstrdup (dos_path);
  info->diw = drive | (1 << 8) | (1 << 11) | (1 << 14);
  info->pos = 0;
  return fd;
}

int DOS_create (const char *path, int attr)
{
  int hindex;
  int fd;

  hindex = get_free_handle_index ();
  ldebug ("(path=`%s', attr=%d) hindex %d\n",
          path, attr, hindex);

  if (-1 == hindex) {
    set_error (TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE);
    return -1;
  }

  fd = case_create (path, attr, file_handles + hindex);

  if (-1 == fd) {
    return -1;
  }
  else {
    return hindex;
  }
}

int DOS_open (const char *path, int dmode)
{
  int hindex;
  int mode;
  int fd;

  hindex = get_free_handle_index ();
  ldebug ("(path=`%s', dmode=%d) hindex %d\n",
          path, dmode, hindex);

  if (-1 == hindex) {
    set_error (TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE);
    return -1;
  }

  if (0 == strcmp (path, "con")) {
    return 1;
  }

  if (EMS_active () && (0 == strcmp (path, "EMMXXXX0"))) {
    struct file_info *info;

    info = file_handles + hindex;
    info->fd = -HEMM;
    info->path = xstrdup (path);
    info->diw = 0;
    info->pos = -1;
    return hindex;
  }

  switch (dmode) {
    case 0:
      mode = O_RDONLY;
      break;

    case 1:
      mode = O_WRONLY;
      break;

    case 2:
      mode = O_RDWR;
      break;

    default:
      set_error (ACCESS_CODE_INVALID);
      return -1;

  }

  fd = case_open (path, mode, hindex);

  if (-1 == fd) {
    return -1;
  }
  else {
    return hindex;
  }
}

#define check_hindex(hindex)                    \
  if (hindex > LENOFA (file_handles)) {         \
    lwarn ("hindex too big %d, max %d\n",       \
           hindex, LENOFA (file_handles));      \
    set_error (INVALID_HANDLE);                 \
    return -1;                                  \
  }

int DOS_get_file_attr (const char *dos_path)
{
  int res;
  int err;
  int drive;
  doserr_e doserr;
  char *unix_path;
  struct stat buf;

  ldebug ("(dos_path=`%s')\n", dos_path);
  doserr = transl (dos_path, &unix_path, &drive);

  if (NO_ERROR != doserr) {
    maybe_free (unix_path);
    set_error (doserr);
    return -1;
  }

  res = lstat (unix_path, &buf);
  err = errno;
  ldebug ("(dos_path=`%s') `%s' res %d\n",
          dos_path, unix_path, res);

  xfree (unix_path);

  if (-1 == res) {
    lwarn ("(dos_path=`%s')\n"
           "unix_path=`%s'\n"
           "syserr %s\n",
           dos_path, unix_path, strerror (err));
    set_error (ACCESS_DENIED);
    return -1;
  }

  if (S_ISDIR (buf.st_mode))
    return DIRECTORY;
  else
    return ARCHIVE;
}

int DOS_dup_handle (int hindex)
{
  int duph;
  int dupfd;

  check_hindex (hindex);

  duph = get_free_handle_index ();
  if (-1 == duph) {
    set_error (TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE);
    return -1;
  }

  dupfd = dup (file_handles[hindex].fd);
  if (-1 == dupfd) {
    int err;

    err = errno;
    lwarn ("(hindex=%d) path `%s'\n"
           "syserr %s\n",
           hindex, file_handles[hindex].path, strerror (err));
    set_error (TOO_MANY_OPEN_FILES_NO_HANDLES_AVAILABLE);
    return -1;
  }

  file_handles[duph].fd = dupfd;
  file_handles[duph].path = xstrdup (file_handles[hindex].path);
  file_handles[duph].diw = file_handles[hindex].diw;
  file_handles[duph].pos = file_handles[hindex].pos;
  return duph;
}

int DOS_ioctl_get_info (int hindex)
{
  int fd;

  check_hindex (hindex);

  fd = file_handles[hindex].fd;

  if (-1 == fd) {
    ldebug ("(hindex=%d) -1\n", hindex);
    return -1;
  }

  ldebug ("(hindex=%d) fd %d name `%s'\n",
          hindex, fd, file_handles[hindex].path);

  switch (fd) {
    case -HCON:
      return 0x82;

    case -HCIN:
      return 0x81;

    case -HEMM:
      return 0x80;

    default:
      return file_handles[hindex].diw;
  }
}

int DOS_read (int hindex, void *buf, int size)
{
  int fd;
  int retval;

  check_hindex (hindex);

  fd = file_handles[hindex].fd;

  ldebug ("(hindex=%d, buf=%p, size=%d)\n",
          hindex, buf, size);

  if (-1 == fd) {
    lwarn ("read from -1 hindex = %d\n", hindex);
    return -1;
  }

  switch (fd) {
    case -HCIN:
      retval = con_read (buf, size);
      break;

    default: {
      static char page[4096];
      int n, t, c;
      char *p;

      p = (char *) buf;
      t = size;
      retval = 0;
      while (t) {
        c = t > 4096 ? 4096 : t;
        n = read (fd, page, c);
        if (0 == n) {
          break;
        }
        if (-1 == n) {
          retval = -1;
          break;
        }
        t -= n;
        memcpy (p, page, n);
        p += n;
        retval += n;
      }

      ldebug ("read %d bytes reval %d\n", size - t, retval);
      break;
    }
      /* retval = (int) read (fd, buf, size); */
  }

  if (-1 == retval) {
    int err;

    err = errno;
    lwarn ("(hindex=%d, buf=%p, size=%d) fd %d\n"
           "name `%s`\n"
           "syserr %s\n",
           hindex, buf, size, fd, file_handles[hindex].path,
           strerror (err));
  }

  file_handles[hindex].pos += retval;
  return retval;
}

int DOS_write (int hindex, void *buf, int size)
{
  int fd;
  int retval;

  check_hindex (hindex);

  ldebug ("(hindex=%d, buf=%p, size=%d)\n",
          hindex, buf, size);

  fd = file_handles[hindex].fd;

  if (-1 == fd) {
    lwarn ("write to -1 hindex = %d\n", hindex);
    return -1;
  }

  switch (fd) {
    case -HCON:
      VGA_print_string_TTY ((char *) buf, size);
      if (con_fd > 0)
        write (con_fd, buf, size);
      return size;

    default:
      break;
  }

  retval = (int) write (fd, buf, size);

  if (-1 == retval) {
    int err;

    err = errno;
    lwarn ("(hindex=%d, buf=%p, size=%d) fd %d\n"
           "name `%s`\n"
           "syserr %s\n",
           hindex, buf, size, fd, file_handles[hindex].path,
           strerror (err));
  }

  file_handles[hindex].diw &= ~(1 << 6);
  file_handles[hindex].pos += retval;

  return retval;
}

int DOS_close (int hindex)
{
  int fd;
  int res;

  check_hindex (hindex);

  fd = file_handles[hindex].fd;

  ldebug ("(hindex=%d) fd %d name `%s'\n",
          hindex, fd, file_handles[hindex].path);

  if (fd < 0) {
    return 0;
  }
  else {
    xfree (file_handles[hindex].path);

    file_handles[hindex].fd = -1;
    file_handles[hindex].path = NULL;

    res = close (fd);
    if (-1 == res) {
      int err;

      err = errno;
      set_error (ACCESS_DENIED);
      lwarn ("close(fd=%d) failed\n"
             "syserr %s\n",
             fd, strerror (err));
      return -1;
    }
    return 0;
  }
}

int DOS_unlink (const char *dos_path)
{
  int res;
  int err;
  int drive;
  doserr_e doserr;
  char *unix_path;

  ldebug ("(dos_path=`%s')\n", dos_path);

  doserr = transl (dos_path, &unix_path, &drive);

  if (NO_ERROR != doserr) {
    maybe_free (unix_path);
    set_error (doserr);
    return -1;
  }

  res = unlink (unix_path);
  err = errno;

  ldebug ("(dos_path=`%s') `%s' res %d\n",
          dos_path, unix_path, res);

  xfree (unix_path);

  if (-1 == res) {
    lwarn ("(dos_path=`%s')\nunix_path=`%s'\n"
           "syserr %s\n",
           dos_path, unix_path, strerror (err));
    set_error (ACCESS_DENIED);
    return -1;
  }

  return 0;
}

int DOS_rename (const char *dos_oldpath, const char *dos_newpath)
{
  int res;
  int err;
  int drive;
  doserr_e doserr;
  char *unix_oldpath;
  char *unix_newpath;

  ldebug ("(dos_oldpath=`%s', dos_newpath=`%s')\n",
          dos_oldpath, dos_newpath);

  doserr = transl (dos_oldpath, &unix_oldpath, &drive);

  if (NO_ERROR != doserr) {
    maybe_free (unix_oldpath);
    set_error (doserr);
    return -1;
  }

  doserr = transl (dos_newpath, &unix_newpath, &drive);
  switch (doserr) {
    case NO_ERROR:
      maybe_free (unix_oldpath);
      maybe_free (unix_newpath);
      set_error (ACCESS_DENIED);
      return -1;

    case PATH_NOT_FOUND:
      maybe_free (unix_oldpath);
      maybe_free (unix_newpath);
      set_error (doserr);
      return -1;

    case FILE_NOT_FOUND:
      break;

    default:
      Aldri ();
      break;
  }

  res = rename (unix_oldpath, unix_newpath);
  err = errno;

  maybe_free (unix_oldpath);
  maybe_free (unix_newpath);

  if (-1 == res) {
    lwarn ("(dos_oldpath=`%s', dos_newpath=`%s')\n"
           "unix_oldpath=`%s'\n"
           "unix_newpath=`%s'\n"
           "syserr %s\n",
           dos_oldpath,
           dos_newpath,
           unix_oldpath,
           unix_newpath,
           strerror (err));

    set_error (ACCESS_DENIED);
    return -1;
  }

  return 0;
}

int DOS_seek (int hindex, int off, int dwhence)
{
  int fd;
  int retval;
  int whence;
  struct file_info *info;

  info = file_handles + hindex;
  check_hindex (hindex);

  ldebug ("(hindex=%d, dwhence=%d, off=%d)\n",
          hindex, dwhence, off);

#if 0
  if ((-1 == off) && (0 == dwhence)) {
    dwhence = 2;
    off = 0;
  }
#endif

  fd = info->fd;

  if (-1 == fd) {
    linfo ("seek from -1 hindex = %d\n", hindex);
    return -0;
  }

  whence = -1;
  switch (dwhence) {
    case 0:
      whence = SEEK_SET;
      break;

    case 1:
      whence = SEEK_CUR;
      break;

    case 2:
      whence = SEEK_END;
      break;

    default: {
      lwarn ("(hindex=%d, whence=%d, off=%d) bogus seek mode\n",
             hindex, dwhence, off);
      set_error (ACCESS_CODE_INVALID);
      return -1;
    }
  }

  if (fd < 0) {
    switch (fd) {
      case -HCON:
        return 0;

      case -HCIN:
        return 0;

      default:
        lwarn ("(hindex=%d, whence=%d, off=%d) special fd %d but not CON\n",
               hindex, dwhence, off, fd);
        return -1;
    }
  }

  retval = lseek (fd, off, whence);
  if (-1 == retval) {
    int err;

    err = errno;
    lwarn ("(hindex=%d, whence=%d, off=%u) fd %d\n"
           "name `%s'\n"
           "syserr %s\n",
           hindex, dwhence, off,
           fd, file_handles[hindex].path,
           strerror (err));
  }

  linfo ("new offset = %d %#x\n", retval, retval);
  file_handles[hindex].pos = retval;
  return retval;
}

void DOS_get_list_of_lists (us *seg, us *off)
{
  DEREF (seg);
  DEREF (off);
  Aldri ();
#if 0
  *seg = LOL >> 4;
  *off = LOL - (*seg * 16);
#endif
}

void DOS_create_child_PSP (ui lin, us next)
{
  DOS_create_PSP (lin >> 4);
  PSP_PARENT_PSP (lin) = PSP_seg;
  PSP_NEXT_SEGMENT (lin) = next;
}

static int env_string_setter (char *string)
{
  char *dup;

  if (NULL == env_strings) {
    nenv_strings_allocated = 2;
    env_strings = malloc (nenv_strings_allocated * sizeof (char *));
    if (NULL == env_strings) {
      log ("failed to allocate initial space for environment strings\n");
      return -1;
    }
  }
  if (nenv_strings_allocated == nenv_strings) {
    nenv_strings_allocated *= 2;
    env_strings = realloc (env_strings,
                           nenv_strings_allocated * sizeof (char *));
    if (NULL == env_strings) {
      log ("failed to allocate %d bytes for environment strings\n",
           nenv_strings_allocated * sizeof (char *));
      return -1;
    }
  }

  dup = strdup (string);
  if (NULL == dup) {
    log ("failed to allocate %d bytes for environment string\n",
         1 + strlen (string));
    return -1;
  }
  env_strings[nenv_strings++] = dup;
  return 0;
}

void DOS_pre_init ()
{
  static struct MUP_key keys[] = {
    {"env.size"   , TAG_IN_INT , {.ii = 16}},
    {"env.string" , TAG_SETTER , {.setter = env_string_setter}},
    {"con.fd"     , TAG_INT    , {.ip = &con_fd}},
  };

  static struct MUP_group group =
    {"dos.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void DOS_init ()
{
  int i;
  ub *mcb;
  ui psp_lin, env_lin;
  us env_paras;

  DEREF (log_blocks);

  log_level = MUP_get_log_level ("dos");

  switch (con_fd) {
    case -1:
    case 1:
    case 2:
      break;

    default:
      Fail ("con.fd = %d can be -1(null), 1(stdout) or 2(stderr)", con_fd);
  }

  for (i = 0; i < MAXDRIVES; i++) {
    char *path;

    path = MUP_get_fs_mapping (i);
    if (path) {
      fs_drives[i].path = xstrdup (path);
      fs_drives[i].cwd = xstrdup (path);
      fs_maped_drives++;
      fs_highest_maped_drive = i;
    }
    else {
      fs_drives[i].path = NULL;
      fs_drives[i].cwd = NULL;
    }
  }

  mcb = ROOT_MCB;
  MCB_ID (mcb) = MCB_LAST_BLOCK;
  MCB_SIZE (mcb) = MEM_USER (MEM_A0000 - (ui) mcb - 16) >> 4;
  MCB_FREE (mcb) = 1;

  /* LOL_FIRST_MCB (LOL) = ((ui) ROOT_MCB) >> 4; */
  /* allocate PSP and environment segments */

  env_paras = MUP_get_int ("dos.env.size");

  if (env_paras <= 0) {
    Fail ("insane env size %d", env_paras);
  }

  if (-1 == DOS_alloc (env_paras, &env_lin)) {
    Fail ("failed to allocate %d bytes for environment, error code %d",
          env_size, dos_errno);
  }

  if (-1 == DOS_alloc (0x10, &psp_lin)) {
    Fail ("failed to allocate 256 bytes for PSP, error code %d",
          dos_errno);
  }

  env_size = env_paras << 4;

  PSP_seg = MEM_USER (psp_lin) >> 4;
  env_seg = MEM_USER (env_lin) >> 4;

  memset ((void *) env_lin, 0, env_size);

  DOS_create_PSP (PSP_seg);

  dta = (char *) (psp_lin + 0x80);

  /* setup files */
  file_handles[0].fd = -HCIN;
  file_handles[0].path = "con";

  file_handles[1].fd = -HCON;
  file_handles[1].path = "con";

  file_handles[2].fd = -HCON;
  file_handles[2].path = "con";

  file_handles[3].fd = -HCON;
  file_handles[3].path = "con";

  file_handles[4].fd = -HCON;
  file_handles[4].path = "con";

  for (i = 5; i < LENOFA (file_handles); i++) {
    file_handles[i].fd = -1;
    file_handles[i].path = NULL;
  }

  for (i = 0; i < nenv_strings; i++) {
    char *env_string;

    env_string = env_strings[i];
    if (NULL == env_string) {
      break;
    }

    DOS_put_env (env_string);
    free (env_string);
  }

  if (env_strings)
    free (env_strings);
}
