#include <unistd.h>
#include <fcntl.h>

#include <stdlib.h>
#include <string.h>

#include <inc/x86.h>
#include <inc/mem.h>
#include <inc/dmm.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/ldt.h>
#include <inc/xdt.h>
#include <inc/sel.h>
#include <inc/dos.h>
#include <inc/red.h>
#include <inc/log.h>
#include <inc/mup.h>
#include <inc/dpmi.h>

#define LOG_PREFIX "I31"
static int log_level = -2;

void I31 (INT_ARGS)
{
  int res, sel;
  char *data;

  if (-2 == log_level)
    log_level = MUP_get_log_level ("i31");

  linfo ("ax = %#06x ", AX (r));

  if (0xff == AH (r)) {
    lwarn ("no handler for %#x\n", AX (r));
    goto error_ok;
  }

  switch (AX (r)) {

    case 0x0000:
      linfo ("ALLOCATE LDT DESCRIPTORS %d\n", CX (r));
      sel = LDT_alloc_selectors (CX (r));
      if (-1 == sel)
        goto error;

      SET_AX (r, sel);
      break;

    case 0x0001:
      linfo ("FREE LDT DESCRIPTOR %#x\n", BX (r));
      LDT_free_selector (BX (r));
      break;

    case 0x0002:
      linfo ("SEGMENT TO DESCRIPTOR %#x\n", BX (r));
      sel = DPMI_create_seg_alias (BX (r));
      if (-1 == sel)
        goto error;

      SET_AX (r, sel);
      break;

    case 0x0003:
      SET_AX (r, 8);
      break;

    case 0x0006:
    {
      ui base;

      if (0 == BX (r))
        goto error_ok;

      base = SEL_get_selector_base (BX (r));
      linfo ("GET SEGMENT BASE ADDRESS %#x = %#x\n",
             BX (r), base);

      base = MEM_USER (base);
      SET_CX (r, GETHIWORD (base));
      SET_DX (r, GETLOWORD (base));
      break;
    }

    case 0x0007:
    {
      ui base;

      base = MKDWORD (CX (r), DX (r));
      linfo ("SET SEGMENT BASE ADDRESS %#x = %#x\n",
             BX (r), base);

      base = MEM_REAL (base);
      LDT_set_selector_base (BX (r), base);
      break;
    }

    case 0x0008:
    {
      ui lim;

      lim = MKDWORD (CX (r), DX (r));

      linfo ("SET SEGMENT LIMIT %#x = %d\n",
             BX (r), lim);

      LDT_set_selector_limit (BX (r), lim);
      break;
    }

    case 0x0009:
      linfo ("SET DESCRIPTOR ACCESS RIGHTS %#x = %#x\n",
             BX (r), CX (r));
      LDT_set_selector_access_rights (BX (r), CX (r));
      break;

    case 0x000a:
      linfo ("CREATE ALIAS DESCRIPTOR %#x\n", BX (r));

      sel = DPMI_create_alias (BX (r));
      if (-1 == sel)
        goto error;

      SET_AX (r, sel);
      break;

    case 0x000b:
    {
      ui base;

      data = (char *) SELLIN (ES (r), EDI (r));
      linfo ("GET DESCRIPTOR %#x, data %p\n", BX (r), data);
      LDT_get_selector_descriptor (BX (r), data);
      base = XDT_get_descriptor_base (data);
      base = MEM_USER (base);
      XDT_set_descriptor_base (data, base);
      break;
    }

    case 0x000c:
    {
      ui base;

      data = (char *) SELLIN (ES (r), EDI (r));
      linfo ("SET DESCRIPTOR %#x, data %p\n", BX (r), data);
      base = XDT_get_descriptor_base (data);
      base = MEM_REAL (base);
      XDT_set_descriptor_base (data, base);
      LDT_set_selector_descriptor (BX (r), data);
      break;
    }

    case 0x0100:
    {
      us sel;
      us seg;

      linfo ("ALLOCATE DOS MEMORY BLOCK %d\n", BX (r));

      if (DPMI_alloc_dos (BX (r), &sel, &seg)) {
        SET_AX (r, 8);
        SET_BX (r, DOS_largest_block ());
        goto error_ok;
      }

      SET_AX (r, seg);
      SET_DX (r, sel);
      break;
    }

    case 0x0101:
      sel = DX (r);

      linfo ("FREE DOS MEMORY BLOCK %#x\n", sel);
      /* workaround for bug in Watcom libc (or codegen) */
      if (SEL_is_system_selector (sel)) {
        lwarn ("attempt to free system selector %#x\n", sel);
        break;
      }

      DPMI_free_dos (sel);
      break;

    case 0x0102:
    {
      int nparas;

      sel = DX (r);
      nparas = BX (r);

      linfo ("RESIZE DOS MEMORY BLOCK %#x, %d\n", sel, nparas);

      if (-1 == DPMI_resize_dos (nparas, sel)) {
        lwarn ("failed to resize dos selector %#x to %d paras\n",
               sel, nparas);

        SET_AX (r, DOS_get_error ());
        SET_BX (r, DOS_largest_block ());
        goto error_ok;
      }
      break;
    }

    case 0x0200:
    {
      us seg, off;

      linfo ("GET REAL MODE INTERRUPT VECTOR %#x\n", BL (r));
      INT_get_rm (BL (r), &seg, &off);
      SET_CX (r, seg);
      SET_DX (r, off);
      break;
    }

    case 0x0201:
      linfo ("SET REAL MODE INTERRUPT VECTOR %#x %04x:%04x\n",
             BL (r), CX (r), DX (r));

      INT_set_rm (BL (r), CX (r), DX (r));
      break;

    case 0x0202:
      linfo ("U GET PROCESSOR EXCEPTION HANDLER VECTOR %#x\n",
             BL (r));
      break;

    case 0x0203:
      linfo ("U SET PROCESSOR EXCEPTION HANDLER VECTOR %#x\n",
             BL (r));
      break;

    case 0x0204:
    {
      us sel;
      ui off;

      linfo ("GET PROTECTED MODE INTERRUPT VECTOR %#x\n",
             BL (r));
      INT_get_pm (BL (r), &sel, &off);

      SET_CX (r, sel);
      EDX (r) = off;
      break;
    }

    case 0x0205:
    {
      us sel;
      ui off;

      sel = CX (r);
      off = EDX (r);
      if ((0 == IS_GDT_SELECTOR (sel))
          && (0 == LDT_is_selector_big (sel))) {
        off &= 0xffff;
      }

      linfo ("SET PROTECTED MODE INTERRUPT VECTOR "
             "%#x %04x:%08x\n",
             BL (r), CX (r), off);

      INT_set_pm (BL (r), CX (r), off);
      break;
    }

    case 0x0300:
    case 0x0301:
    case 0x0302:
    {
      int nint;
      data = (char *) SELLIN (ES (r), EDI (r));

      if (0x0300 == AX (r)) {
        nint = BL (r);

        linfo ("SIMULATE REAL MODE INTERRUPT %#x %p %d\n",
               nint, data, CX (r));
      }
      else {
        nint = -1;
        linfo ("CALL REAL MODE PROCEDURE WITH %s FRAME %04x:%04x %04x:%04x %d\n",
               AL (r) & 1 ? "FAR" : "IRET",
               DPMI_CS (data), DPMI_IP (data),
               DPMI_SS (data), DPMI_SP (data),
               CX (r));
      }
      DPMI_call_rm (nint, AL (r) & 1, CX (r), data);
      break;
    }

    case 0x0303:
    {
      int ncallback;

      linfo ("ALLOCATE REAL MODE CALLBACK ADDRESS %04x:%08x %04x:%08x\n",
             DS (r), ESI (r), ES (r), EDI (r));

      ncallback = DPMI_alloc_callback (DS (r), ESI (r), ES (r), EDI (r));
      if (-1 == ncallback)
        goto error_ok;

      SET_CX (r, RED_SEGMENT);
      SET_DX (r, RED_CALLBACK_0 + ncallback);
      break;
    }

    case 0x0304:
    {
      int ncallback;

      ncallback = SEGLIN (CX (r), DX (r)) - RED_FLAT (RED_CALLBACK_0);

      linfo ("FREE REAL MODE CALLBACK ADDRESS %04x:%04x = %d\n",
             CX (r), DX (r), ncallback);

      res = DPMI_free_callback (ncallback);
      if (-1 == res)
        goto error_ok;
      break;
    }

    case 0x0305:
      linfo ("GET STATE SAVE/RESTORE ADDRESSES\n");
      SET_AX (r, 0);
      SET_BX (r, RED_SEGMENT);
      SET_CX (r, RED_SAVE_RESTORE_RM);
      SET_SI (r, SEL_get_user_code ());
      EDI (r) = RED_FLAT (RED_SAVE_RESTORE_PM);
      break;

    case 0x0306:
      linfo ("GET RAW MODE SWITCH ADDRESSES\n");
      SET_BX (r, RED_SEGMENT);
      SET_CX (r, RED_RAW_SWITCH_TO_PM);
      SET_SI (r, SEL_get_user_code ());
      EDI (r) = RED_FLAT (RED_RAW_SWITCH_TO_RM);
      break;

    case 0x0400:
      linfo ("GET DPMI VERSION\n");
      SET_AX (r, 0x0009);
      SET_BX (r, 1);
      SET_CL (r, 5);
      SET_DX (r, 0x0870);
      break;

    case 0x0500:
    {
      ui *p;

      p = (ui *) SELLIN (ES (r), EDI (r));
      linfo ("GET FREE MEMORY INFORMATION %p\n", (void *) p);
      p[0] = DMM_largest ();
      p[1] = p[0] >> 12;
      p[2] = p[1];
      p[3] = (MEM_ram_size () - 0x100000) >> 12;
      p[4] = p[3];
      p[5] = p[1];
      p[6] = p[3];
      p[7] = p[1];
      p[8] = 0;
      break;
    }

    case 0x0501:
    {
      ui size;
      ui addr, handle;

      size = MKDWORD (BX (r), CX (r));
      linfo ("ALLOCATE MEMORY BLOCK %d\n", size);
      res = DPMI_alloc (size, &handle, &addr);

      if (-1 == res)
        goto error_ok;

      SET_BX (r, GETHIWORD (addr));
      SET_CX (r, GETLOWORD (addr));
      SET_SI (r, GETHIWORD (handle));
      SET_DI (r, GETLOWORD (handle));
      break;
    }

    case 0x0502:
    {
      ui handle;

      handle = MKDWORD (SI (r), DI (r));
      linfo ("FREE MEMORY BLOCK %#x\n", handle);

      res = DPMI_free (handle);
      if (-1 == res)
        goto error_ok;

      break;
    }

    case 0x0503:
    {
      ui new_size;
      ui handle, old_handle;
      ui addr;

      new_size = MKDWORD (BX (r), CX (r));
      old_handle = MKDWORD (SI (r), DI (r));

      linfo ("RESIZE MEMORY BLOCK %#x %d\n", handle, new_size);

      res = DPMI_realloc (old_handle, new_size, &handle, &addr);
      if (-1 == res)
        goto error_ok;

      SET_BX (r, GETHIWORD (addr));
      SET_CX (r, GETLOWORD (addr));
      SET_SI (r, GETHIWORD (handle));
      SET_DI (r, GETLOWORD (handle));
      break;
    }

    case 0x0506:
      lwarn ("U GET PAGE ATTRIBUTES %#x %d %d %04x:%08x\n",
             ESI (r),
             EBX (r),
             ECX (r),
             ES (r), EDX (r));
      goto error_ok;

    case 0x0600:
      linfo ("LOCK LINEAR REGION %#x %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0601:
      linfo ("UNLOCK LINEAR REGION %#x %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0604:
      linfo ("GET PAGE SIZE\n");
      SET_BX (r, 0);
      SET_CX (r, 0x1000);
      break;

    case 0x0700:
      linfo ("U MARK PAGES AS PAGING CANDIDATES %d - %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0701:
      linfo ("U DISCARD PAGES %d - %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0702:
      linfo ("U MARK PAGES AS PAGING CANDIDATES %d - %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0703:
      linfo ("U DISCARD PAGE CONTENTS %d - %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0800:
      linfo ("U PHYSICAL ADDRESS MAPPING %#x %d\n",
             MKDWORD (BX (r), CX (r)),
             MKDWORD (SI (r), DI (r)));
      break;

    case 0x0801:
      linfo ("U FREE PHYSICAL ADDRESS MAPPING %#x\n",
             MKDWORD (BX (r), CX (r)));
      break;

    case 0x0900:
    case 0x0901:
    case 0x0902:
      res = DPMI_vif (AL (r));
      linfo ("GET %s%s VIRTUAL INTERRUPT STATE %d\n",
             (AL (r) < 2) ? "AND " : "",
             (AL (r) < 2) ? (AL (r) ? "ENABLE" : "DISABLE") : "",
             res);
      SET_AL (r, res);
      break;

    case 0x0a00:
      data = (char *) SELLIN (DS (r), ESI (r));
      linfo ("GET VENDOR SPECIFIC API ENTRY POINT %s\n", data);
      goto error_ok;

    default:
      UNHANDLED_INT (r);
  }

  CLR_CF (r);
  return;

 error:
  SET_CF (r);
  UNHANDLED_INT (r);

 error_ok:
  lwarn ("function %#x completed with error\n", AX (r));
  SET_CF (r);
}
