#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#include <inc/dos.h>
#include <inc/mem.h>
#include <inc/ldr.h>
#include <inc/sel.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "LDR"
static int log_level;

#define DOS_SIG(h)        (*(us*)(h+0x00))
#define DOS_LPG(h)        (*(us*)(h+0x02))
#define DOS_NPG(h)        (*(us*)(h+0x04))
#define DOS_NRE(h)        (*(us*)(h+0x06))
#define DOS_HSZ(h)        (*(us*)(h+0x08))
#define DOS_MIN(h)        (*(us*)(h+0x0a))
#define DOS_MAX(h)        (*(us*)(h+0x0c))
#define DOS_ISS(h)        (*(us*)(h+0x0e))
#define DOS_ISP(h)        (*(us*)(h+0x10))
#define DOS_SUM(h)        (*(us*)(h+0x12))
#define DOS_IIP(h)        (*(us*)(h+0x14))
#define DOS_ICS(h)        (*(us*)(h+0x16))
#define DOS_ROF(h)        (*(us*)(h+0x18))
#define DOS_NOV(h)        (*(us*)(h+0x1a))

void LDR_load (const char *path, struct imginfo *info)
{
  int i;
  int com;
  int fd;
  char *h;
  us *rel;
  int map_len;
  ui base, entry, stack, psp;
  ui base_seg;
  ui hdr_size, res_size;
  ui img_size;
  ui res_paras, min_paras, alloc_paras, largest;

  DEREF (entry);
  DEREF (stack);

  hdr_size = 0xffffffffU;
  /* res_size = 0xffffffffU; */

  log_level = MUP_get_log_level ("ldr");

  linfo ("path=`%s'\n", path);

  fd = open (path, O_RDONLY);
  if (-1 == fd) {
    ERRFail ("open `%s'", path);
  }

  map_len = lseek (fd, 0, SEEK_END);
  if (-1 == map_len) {
    ERRFail ("lseek `%s':%d", path, fd);
  }

  h = mmap (0, map_len, PROT_READ, MAP_PRIVATE, fd, 0);

  if (MAP_FAILED == h) {
    ERRFail ("mmap %d for `%s':%d", map_len, path, fd);
  }

  if ((0x5a4d != DOS_SIG (h)) && (0x4d5a != DOS_SIG (h))) {
    if (map_len >= (0x10000 - 0x100 - 2)) {
      Fail ("`%s' is not DOS executable sig %#x nor com", path, DOS_SIG (h));
    }
    com = 1;
    alloc_paras = DOS_largest_block () + 0x10;
    img_size = map_len;
  }
  else {
    ui est_size;

    com = 0;
    hdr_size = DOS_HSZ (h) * 16;
    img_size = (DOS_NPG (h) - (DOS_LPG (h) != 0)) * 512 + DOS_LPG (h);
    est_size = DOS_ISS (h) * 16 + DOS_ISP (h) + 0x100;
    res_size = img_size - hdr_size;

    if (res_size < est_size)
      res_size = est_size;

    res_paras = ((res_size + 15) & ~15) / 16;
    min_paras = res_paras; /* + DOS_MIN (h); */
    largest = DOS_largest_block () + 0x10;

    if (largest < min_paras)
      Fail ("not enough low-memory need %d paragraphs got only %d",
            min_paras, largest);

    if (0xffff == DOS_MAX (h)) {
      alloc_paras = largest;
    }
    else {
      ui max_paras, paras;

      max_paras = res_paras + DOS_MAX (h);
      paras = min_paras > max_paras ? min_paras : max_paras;

      if (paras > largest)
        alloc_paras = largest;
      else
        alloc_paras = paras;
    }
  }
  psp = DOS_get_PSP ();
  base = psp;

  if (-1 == DOS_resize (base, alloc_paras)) {
    Fail ("failed to allocate %d paragraphs of low-memory", alloc_paras);
  }

  *(us *) (psp + 2) = ((MEM_USER (psp) >> 4) + alloc_paras + 1);

  if (com) {
    base_seg = MEM_USER (base) >> 4;
    memcpy ((void *) (base + 0x100), h, map_len);
    info->cs = base_seg;
    info->ip = 0x100;
    info->ss = base_seg;
    info->sp = 0xfffe;
    *(us *) base = 0x20cd;
    *(us *) (base + 0xfffe) = 0;
    entry = info->cs * 16 + 0x100;
    stack = info->ss * 16 + info->sp;
  }
  else {
    base += 0x100;
    base_seg = MEM_USER (base) >> 4;

    memcpy ((void *) base, h + hdr_size, img_size - hdr_size);

    rel = (us *) (h + DOS_ROF (h));
    for (i = 0; i < DOS_NRE (h); i++, rel += 2) {
      ui lin;

      lin = base + rel[1] * 16 + rel[0];
      *(us *) lin += base_seg;
    }

    entry = DOS_ICS (h) * 16 + DOS_IIP (h) + base;
    stack = DOS_ISS (h) * 16 + DOS_ISP (h) + base;

    info->cs = base_seg + DOS_ICS (h);
    info->ip = DOS_IIP (h);
    info->ss = base_seg + DOS_ISS (h);
    info->sp = DOS_ISP (h);
  }

  /* BPT; */
  ldebug ("base %#x %#x\n", base, base_seg);
  ldebug ("cs:ip %04x:%04x %#x %04x:%04x\n",
          DOS_ICS (h), DOS_IIP (h), entry, info->cs, info->ip);
  ldebug ("ss:sp %04x:%04x %#x %04x:%04x\n",
          DOS_ISS (h), DOS_ISP (h), stack, info->ss, info->sp);

  if (-1 == munmap (h, map_len)) {
    ERRFail ("munmap for %p:%d `%s':%d", h, map_len, path, fd);
  }

  if (-1 == close (fd)) {
    ERRFail ("close for `%s':%d", path, fd);
  }
}
