#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include <inc/sig.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_BUF_SZ 16384

static char *logname;
static const char *prev_func;
static int pref_len;
static int prev_cr;
static const char *prev_pref;
static int log_fd = -1;
static int log_aux_fd = 2;
static char log_buf[LOG_BUF_SZ];
static char pfiller[] = "        | ";
#define FILSZ (sizeof (pfiller) - 4)
#define BARSZ 3

static void logwrite (const char *msg, int n)
{
  if (n > 0) {
    if (-1 != log_fd)
      write (log_fd, msg, n);
    if (log_aux_fd > 0)
      write (log_aux_fd, msg, n);
  }
}

void LOG_log (const char *fmt, ...)
{
  int n;
  int was_blocked;
  va_list ap;

  was_blocked = SIG_block ();
  va_start (ap, fmt);
  n = vsnprintf (log_buf, LOG_BUF_SZ, fmt, ap);
  logwrite (log_buf, n);
  va_end (ap);

  SIG_unblock (was_blocked);
}

static void log_al ()
{
  unsigned int i;

  for (i = 0; i < (pref_len - 1) / FILSZ; i++) {
    logwrite (pfiller, FILSZ);
  }

  i = (pref_len - 1) % FILSZ;
  logwrite (pfiller + FILSZ - i, i + BARSZ);
}

void LOG_log_pref (const char *pref,
                   const char *func,
                   const char *fmt, ...)
{
  int n;
  int was_blocked;
  va_list ap;

  was_blocked = SIG_block ();

  if ((prev_pref == pref) && (prev_func == func)) {
    if (prev_cr)
      log_al ();
  }
  else {
    int func_len;
    prev_pref = pref;
    prev_func = func;
    pref_len = strlen (pref);
    logwrite (pref, pref_len);
    logwrite (": ", 2);
    func_len = strlen (func);
    if ((func_len != pref_len) || (0 != strcasecmp (func, pref))) {
      logwrite (func, func_len);
      logwrite (" ", 1);
    }
  }

  va_start (ap, fmt);
  n = vsnprintf (log_buf, 16384, fmt, ap);
  prev_cr = 0;
  if (n > 0) {
    char *s;
    char *ns;

    s = log_buf;
    do {
      int len;

      ns = strchr (s, '\n');
      prev_cr = NULL != ns;

      if (ns) {
        len = ns - s + 1;
      }
      else {
        len = strlen (s);
      }

      if (ns && (0 == ns[1])) {
        ns = 0;
      }

      logwrite (s, len);

      if (ns) {
        log_al ();
      }

      s += len;
    } while (ns);
  }
  va_end (ap);

  SIG_unblock (was_blocked);
}

void LOG_fail (const char *func, const char *file, int line,
               const char *fmt, ...)
{
  int n;
  va_list ap;

  log_aux_fd = 2;
  LOG_log ("Fail in %s at %s:%d\n", func, file, line);

  SIG_block ();

  va_start (ap, fmt);
  n = vsnprintf (log_buf, 16384, fmt, ap);
  logwrite (log_buf, n);
  va_end (ap);

  logwrite ("\n", 1);
  SIG_break (1);
  exit (EXIT_FAILURE);
}

void LOG_err_fail (const char *func, const char *file, int line,
                   const char *fmt, ...)
{
  int n;
  va_list ap;
  int err;

  err = errno;
  log_aux_fd = 2;
  LOG_log ("Fail in %s at %s:%d\n", func, file, line);

  SIG_block ();

  va_start (ap, fmt);
  n = vsnprintf (log_buf, 16384, fmt, ap);
  logwrite (log_buf, n);
  va_end (ap);

  logwrite ("\n", 1);

  LOG_log ("error: %s\n", strerror (err));
  SIG_break (1);
  exit (EXIT_FAILURE);
}

void LOG_pre_init ()
{
  static struct MUP_key keys[] = {
   {"all"               , TAG_LOG_LEVEL , {.ll = -1}},
   {"file"              , TAG_STRING    , {.sp = &logname}},
   {"fd"                , TAG_INT       , {.ip = &log_aux_fd}},
   #include <inc/lce.inc>
  };

  static struct MUP_group group =
    {"log.", LENOFA (keys), keys, NULL};

  MUP_add_group (&group);
}

void LOG_init ()
{
  switch (log_aux_fd) {
    case -1:
    case 1:
    case 2:
      break;

    default:
      fprintf (stderr, "log.fd = %d, can be -1(nothing), 1 or 2\n", log_aux_fd);
      exit (EXIT_FAILURE);
  }

  if (logname) {
    log_fd = open (logname, O_CREAT | O_WRONLY | O_TRUNC, 0666);
    if (-1 == log_fd) {
      int err;

      err = errno;
      fprintf (stderr, "open log `%s': %s\n", logname, strerror (err));
      exit (EXIT_FAILURE);
    }
  }
}
