#include <string.h>
#include <stdlib.h>

#include <inc/x86.h>
#include <inc/pie.h>
#include <inc/ldt.h>
#include <inc/cli.h>
#include <inc/eio.h>
#include <inc/asm.h>
#include <inc/sel.h>
#include <inc/sig.h>
#include <inc/int.h>
#include <inc/log.h>
#include <inc/mup.h>
#include <inc/dpmi.h>

static int log_level;
#define LOG_PREFIX "PIE"

static int watch_cli = -1;
static enum {Never, IfEsp, IfEspAndFpu, Always} make_stack_big = Never;
static cli_type_e cli_type;
static int *big_sels;
static int nbig_sels;

static void outp (int nport, void *src, int size)
{
  ui val;

  switch (size) {
    case 1: val = *(ub *) src; break;
    case 2: val = *(us *) src; break;
    case 4: val = *(ui *) src; break;
    default:
      Aldri ();
  }
  EIO_write (nport, size, val);
}

static void inp (int nport, void *dst, int size)
{
  ui val;

  val = EIO_read (nport, size);
  switch (size) {
    case 1: *(ub *) dst = val; break;
    case 2: *(us *) dst = val; break;
    case 4: *(ui *) dst = val; break;
  }
}

static void emu (int real_mode, struct vm86_regs *r)
{
  int dmode, amode, ilen, psize, paddr, big, override;
  ui amask, dmask, opc, prefix;
  ub *pc;

  ilen = 0;
  prefix = 0;
  big = 0;
  dmode = 0;
  amode = 0;
  override = -1;

  if (real_mode) {
    pc = (ub *) ((r->cs << 4) + GETLOWORD (r->eip));
  }
  else {
    big = LDT_is_selector_big (r->cs);
    pc = (ub *) SEL_get_selector_base (r->cs);

    if (big) {
      dmode = 1;
      amode = 1;
      pc += r->eip;
    }
    else
      pc += GETLOWORD (r->eip);
  }

  ilen = X86_process_prefixes (pc, &prefix, &override);

  if (-1 != override) {
    switch (override) {
      case X86_PREFIX_ES: override = r->es; break;
      case X86_PREFIX_CS: override = r->cs; break;
      case X86_PREFIX_SS: override = r->ss; break;
      case X86_PREFIX_DS: override = r->ds; break;
      case X86_PREFIX_FS: override = r->fs; break;
      case X86_PREFIX_GS: override = r->gs; break;
      default:
        Aldri ();
    }
  }

  opc = pc[ilen++];

  if (prefix & X86_PREFIX_DATA) dmode ^= 1;
  if (prefix & X86_PREFIX_ADDR) amode ^= 1;

  dmask = 0xffff;
  amask = 0xffff;

  if (dmode) dmask |= 0xffff0000;
  if (amode) amask |= 0xffff0000;

  psize = (0 == (opc & 1)) ? 1 : (dmode ? 4 : 2);
  paddr = GETLOWORD (r->edx);

  if (0x0f == opc) opc = 0x100 | pc[ilen++];

  switch (opc) {
    case 0xe4:                /* in al, Ib */
    case 0xe5:                /* in eAX, Ib */
      paddr = pc[ilen++];
    case 0xec:                /* in al, dx */
    case 0xed:                /* in eAX, dx */
      inp (paddr, &r->eax, psize);
      break;

    case 0xe6:                /* out Ib, al */
    case 0xe7:                /* out Ib, eAX */
      paddr = pc[ilen++];
    case 0xee:                /* out dx, al */
    case 0xef:                /* out dx, eAX */
      outp (paddr, &r->eax, psize);
      break;

    case 0x6c:                  /* ins Xb */
    case 0x6d:                  /* ins Xv */
    case 0x6e:                  /* outs Xb */
    case 0x6f:                  /* outs Xv */
    {
      ub *base;
      int incr, sel;
      ui cx, xi, i;
      long *reg;

      if (opc < 0x6e) {
        sel = r->es;
        reg = &r->edi;
      }
      else {
        sel = r->ds;
        reg = &r->esi;
      }

      xi = *reg & amask;
      if (-1 != override) sel = override;

      incr = (r->eflags & X86_EFLAGS_DF) ? -psize : psize;
      if (real_mode) {
        base = (ub *) (sel << 4);
      }
      else {
        base = (ub *) LDT_get_selector_base (sel);
      }

      if (prefix & (X86_PREFIX_REP | X86_PREFIX_REPNE))
        cx = r->ecx & amask;
      else
        cx = 1;

      for (i = 0; i < cx; i++) {
        if (opc < 0x6e)
          inp (paddr, base + xi, psize);
        else
          outp (paddr, base + xi, psize);
        xi += incr;
        xi &= amask;
      }

      if (prefix & (X86_PREFIX_REP | X86_PREFIX_REPNE))
        r->ecx &= ~amask;
      *reg = (*reg & ~amask) | xi;
      break;
    }

    case 0x120:
      if (0xc0 == pc[ilen++]) { /* mov eax, cr0 - pmode/w */
        r->eax = 0x00000001;
        break;
      }

    case 0x07:                  /* pop es */
    {
      us *sp, sbig;

      if (real_mode)
        goto fail;

      sbig = LDT_is_selector_big (r->ss);
      sp = (us *) SELLIN (r->ss, r->esp);

      if (0x40 == *sp) {      /* fake bios selector */
        if (sbig) {
          r->esp += (2 << dmode);
        }
        else {
          r->esp = (r->esp & ~0xffff) | (GETLOWORD (r->esp) + (2 << dmode));
        }

        r->es = DPMI_get_bios_selector ();
        break;
      }
      goto fail;
    }

    case 0x8e:
    {
      ub modrm;
      us sel, is_bios;

      if (real_mode)
        goto fail;

      modrm = pc[ilen++];
      if (0xc0 != (modrm & 0xc0))
        goto fail;

      is_bios = 0;
      switch (modrm & 7) {
        case 0: is_bios = GETLOWORD (r->eax) == 0x40; break;
        case 1: is_bios = GETLOWORD (r->ecx) == 0x40; break;
        case 2: is_bios = GETLOWORD (r->edx) == 0x40; break;
        case 3: is_bios = GETLOWORD (r->ebx) == 0x40; break;
        case 4: is_bios = GETLOWORD (r->esp) == 0x40; break;
        case 5: is_bios = GETLOWORD (r->ebp) == 0x40; break;
        case 6: is_bios = GETLOWORD (r->esi) == 0x40; break;
        case 7: is_bios = GETLOWORD (r->edi) == 0x40; break;
      }

      if (0 == is_bios)
        goto fail;

      sel = DPMI_get_bios_selector ();

      switch ((modrm >> 3) & 7) {
        case 0: r->es = sel; break;
        case 2: r->ss = sel; break;
        case 3: r->ds = sel; break;
        case 4: r->fs = sel; break;
        case 5: r->gs = sel; break;
        default:
          goto fail;
      }
      break;
    }

  fail:
    default:
      Fail ("can not emulate instruction at %#06x %#010lx %p",
            r->cs, r->eip, pc);
  }

  if (real_mode || (0 == big)) {
    r->eip = (r->eip & ~0xffff) | (GETLOWORD (r->eip) + ilen);
  }
  else {
    r->eip += ilen;
  }
}

void PIE_emu16 (struct vm86_regs *r)
{
  emu (1, r);
}

void PIE_emu (regs86 *r)
{
  ub *pc;
  int incr;

  incr = 1;
  pc = (ub *) SELLIN (SIGCS (r), SIGEIP (r));

  switch (*pc) {
    case 0xfa:                  /* cli */
    {
      int enabled, in_pic;

      enabled = INT_is_enabled ();
      in_pic = SIG_is_in_pic_loop ();

      linfo ("cli at %#06x %#010x\n"
             "enabled %d pic loop %d watch_cli %d\n",
             (int) SIGCS (r), (int) SIGEIP (r), enabled, in_pic, watch_cli);

      if ((CLI_IGNORE != cli_type) && (0 == in_pic)) {
        if (CLI_PUSHF == cli_type) {
          int n;

          n = CLI_find_pushf (enabled, r);

          if (-1 != n) {
            linfo ("found pushf %d %p\n", n, pc);
          }
          else {
            linfo ("pushf not found %#06x %#010x %p\n",
                   SIGCS (r), (int) SIGEIP (r), pc);
          }
          if (enabled && (-1 == watch_cli))
            watch_cli = n;
        }

        if (enabled)
          INT_disable ();
      }
      break;
    }

    case 0xfb:                  /* sti */
    {
      int enabled, in_pic;

      enabled = INT_is_enabled ();
      in_pic = SIG_is_in_pic_loop ();

      linfo ("sti at %#x enabled %d pic loop %d watch_cli %d\n",
             (int) SIGEIP (r), enabled, in_pic, watch_cli);

      if (0 == in_pic)
        INT_enable ();
      break;
    }

    case 0xcd:                  /* int Ib */
      SIGEIP (r) += 2;
      SIGEFL (r) &= ~X86_EFLAGS_TF;
      INT_simulate (pc[1], r);
      incr = 0;
      break;

    default:
    {
      struct vm86_regs d;

      memset (&d, 0, sizeof (d));
      d.eax = SIGEAX (r);
      d.ebx = SIGEBX (r);
      d.ecx = SIGECX (r);
      d.edx = SIGEDX (r);
      d.esi = SIGESI (r);
      d.edi = SIGEDI (r);
      d.ebp = SIGEBP (r);
      d.esp = SIGESP (r);

      d.cs = SIGCS (r);
      d.ds = SIGDS (r);
      d.es = SIGES (r);
      d.ss = SIGSS (r);
      d.fs = SIGFS (r);
      d.gs = SIGGS (r);

      d.eip = SIGEIP (r);
      d.eflags = SIGEFL (r);
      emu (0, &d);
      SIGEAX (r) = d.eax;
      SIGEBX (r) = d.ebx;
      SIGECX (r) = d.ecx;
      SIGEDX (r) = d.edx;
      SIGESI (r) = d.esi;
      SIGEDI (r) = d.edi;
      SIGEBP (r) = d.ebp;
      SIGESP (r) = d.esp;

      SIGCS (r) = d.cs;
      SIGDS (r) = d.ds;
      SIGES (r) = d.es;
      SIGSS (r) = d.ss;
      SIGFS (r) = d.fs;
      SIGGS (r) = d.gs;

      SIGEIP (r) = d.eip;
      SIGEFL (r) = d.eflags;
      return;
    }
  }
  SIGEIP (r) += incr;
}

int PIE_trap (regs86 *r)
{
  int enabled;

  enabled = INT_is_enabled ();

  linfo ("watch_cli = %d enabled = %d\n", watch_cli, enabled);

/*   if ((-1 == watch_cli) || INT_is_enabled ()) { */
/*     lwarn ("watch_cli = %d enabled = %d\n", watch_cli, enabled); */
/*     return -1; */
/*   } */

  if (-1 == watch_cli) {
    lwarn ("watch_cli = %d enabled = %d\n", watch_cli, enabled);
    return -1;
  }

  SIGEFL (r) &= ~X86_EFLAGS_TF;
  /* watch_cli = -1; */
  lverbose ("reenabling pushfed\n");
  INT_enable ();
  return 0;
}

static void add_big_sel (int sel)
{
  if (NULL == big_sels) {
    big_sels = xmalloc (1);
  }
  else {
    big_sels = xrealloc (big_sels, nbig_sels + 1);
  }

  big_sels[nbig_sels] = sel;
  nbig_sels += 1;
}

static int is_ex_small_sel (int sel)
{
  int i;

  for (i = 0; i < nbig_sels; i++)
    if (sel == big_sels[i])
      return 1;
  return 0;
}

int PIE_stack_hack (regs86 *r)
{
  int dmode, amode, override, c32, s32, ilen, fp, ex_small;
  ui prefix;
  ui opc;
  ub modrm, mod, rm, base;
  reg86 *regp;
  ub *pc;

  c32 = LDT_is_selector_big (SIGCS (r));
  s32 = LDT_is_selector_big (SIGSS (r));

  ex_small = 0;
  if (s32) {
    if (0 == is_ex_small_sel (SIGSS (r))) {
      return -1;
    }
    else {
      ex_small = 1;
    }
  }

  prefix = 0;
  dmode = 0;
  amode = 0;
  ilen = 0;
  override = -1;

  pc = (ub *) SEL_get_selector_base (SIGCS (r));

  if (c32) {
    dmode = 1;
    amode = 1;
    pc += SIGEIP (r);
  }
  else
    pc += GETLOWORD (SIGEIP (r));

  ilen = X86_process_prefixes (pc, &prefix, &override);

  if (prefix & X86_PREFIX_DATA) dmode ^= 1;
  if (prefix & X86_PREFIX_ADDR) amode ^= 1;

  if (0 == amode) {
    if (ex_small) {
      SIGESP (r) &= 0xffff;
      return 1;
    }
    return -1;
  }

  opc = pc[ilen++];
  if (0xf == opc) opc = 0x100 | pc[ilen++];

  fp = 0;
  switch (opc) {
    case 0x0d8:
    case 0x0d9:
    case 0x0da:
    case 0x0db:
    case 0x0dc:
    case 0x0dd:
    case 0x0de:
    case 0x0df:
      fp = 1;
      ilen++;
      break;

    case 0x006:
    case 0x007:
    case 0x00e:
    case 0x016:
    case 0x017:
    case 0x01e:
    case 0x01f:

    case 0x050:
    case 0x051:
    case 0x052:
    case 0x053:
    case 0x054:
    case 0x055:
    case 0x056:
    case 0x057:

    case 0x058:
    case 0x059:
    case 0x05a:
    case 0x05b:
    case 0x05c:
    case 0x05d:
    case 0x05e:
    case 0x05f:
    case 0x060:
    case 0x061:
    case 0x068:
    case 0x06a:
    case 0x09a:
    case 0x09c:
    case 0x09d:
    case 0x0c2:
    case 0x0c3:
    case 0x0c8:
    case 0x0c9:
    case 0x0cc:
    case 0x0cd:
    case 0x0ce:
    case 0x0cf:
    case 0x0e8:

    case 0x1a0:
    case 0x1a1:
    case 0x1a8:
    case 0x1a9:
      if (ex_small) {
        SIGESP (r) &= 0xffff;
        return 1;
      }
      return -1;
  }

  modrm = pc[ilen++];
  if (0xc0 == (modrm & 0xc0))
    return -1;

  mod = (modrm >> 6) & 3;
  rm = modrm & 7;

  if (4 == rm) {
    ub sib;

    sib = pc[ilen++];
    base = sib & 7;
  }
  else {
    base = rm;
  }

  if (5 == (base | -mod))
    return -1;

  if ((-1 == override) && (4 == (base & 6)))
    override = X86_PREFIX_SS;

  if (X86_PREFIX_SS != override)
    return -1;

  regp = SIG_gregp (r, base);

  if ((*regp & ~0xffff) == (SIGESP (r) & ~0xffff)) {
    *regp &= 0xffff;
  }
  else
    return -1;

  switch (make_stack_big) {
    /* remove me once FPU in x86 and turtle is in place */
    case Never:
      break;

    case IfEsp:
      if (4 == base) {
        LDT_set_selector_access_rights (SIGSS (r), 0x40f3);
        add_big_sel (SIGSS (r));
      }
      break;

    case IfEspAndFpu:
      if ((4 == base) && (fp)) {
        LDT_set_selector_access_rights (SIGSS (r), 0x40f3);
        add_big_sel (SIGSS (r));
      }
      break;

    case Always:
      LDT_set_selector_access_rights (SIGSS (r), 0x40f3);
      add_big_sel (SIGSS (r));
  }
/*   if ((4 == base) && (make_stack_big) && (fp)) { */
/*     LDT_set_selector_access_rights (SIGSS (r), 0x40f3); */
/*     return 1; */
/*   } */

  return 1;
}

static int sbig_setter (char *desc)
{
  if (0 == strcmp (desc, "always"))
    make_stack_big = Always;
  else if (0 == strcmp (desc, "never"))
    make_stack_big = Never;
  else if (0 == strcmp (desc, "esp"))
    make_stack_big = IfEsp;
  else if (0 == strcmp (desc, "esp.and.fpu"))
    make_stack_big = IfEspAndFpu;
  else {
    log ("bogus make stack big type must be one of:\n"
         "`always', `never', `esp', `esp.and.fpu'\n");
    return -1;
  }
  return 1;
}

void PIE_pre_init ()
{
  static struct MUP_key keys[] =
    {{"make.stack.big", TAG_SETTER, {.setter = &sbig_setter}}};

  MUP_add_keys (keys, LENOFA (keys));
}

void PIE_init ()
{
  log_level = MUP_get_log_level ("pie");

  cli_type = CLI_get_type ();
}
