#include <inc/eio.h>
#include <inc/pit.h>
#include <inc/pic.h>
#include <inc/int.h>
#include <inc/mem.h>
#include <inc/tid.h>
#include <inc/log.h>
#include <inc/mup.h>

#define LOG_PREFIX "PIT"
static int log_level;

#define PIT_HZ 1193182ULL

enum ACCESSMODE {
  AccessHiByte = 1,
  AccessLoByte = 2,
  AccessLoHi = 3
};

#define Channel(ch) (((ch)->val >> 6) & 3)
#define AccessMode(ch) (((ch)->val >> 4) & 3)
#define Mode(ch) (((ch)->val >> 1) & 7)
#define BCD(ch) ((ch)->val & 1)

#define GetFlipFlop(ch) ((ch)->status & 1)
#define ClrFlipFlop(ch) ((ch)->status &= ~1)

#define WasWritten(ch) (0 != ((ch)->status & 2))
#define SetWritten(ch) ((ch)->status |= 2)
#define ClrWritten(ch) ((ch)->status &= ~2)

#define WasTouched(ch) (0 != ((ch)->status & 4))
#define SetTouched(ch) ((ch)->status |= 4)
#define ClrTouched(ch) ((ch)->status &= ~4)

#define WasLatched(ch) (0 != ((ch)->status & 8))
#define SetLatched(ch) ((ch)->status |= 8)
#define ClrLatched(ch) ((ch)->status &= ~8)

#define GetOutPin(ch) (0 != ((ch)->status & 16))
#define SetOutPin(ch) ((ch)->status |= 16)
#define ClrOutPin(ch) ((ch)->status &= ~16)

struct ctc_chan {
  ub val;
  ub status;
  us reload;
  us new_reload;
  us latch;
  ul base;
  ul counter;
  ul touched_at;
};

static us mode3_default_latch;
static struct ctc_chan channels[3];
static const char *access_names[] =
{ "latch", "lobyte", "hibyte", "lo/hibyte" };

static void log_chan (const char *cap, struct ctc_chan *ch)
{
  DEREF (cap);
  DEREF (ch);
  lverbose ("%s\n"
            "mode %d access %d status %#x reload %d latch %d\n"
            "flipflop %d outpin %d was(wtl): %03x\n",
            cap,
            Mode (ch), AccessMode (ch), ch->status, ch->reload, ch->latch,
            GetFlipFlop (ch), GetOutPin (ch),
            (WasWritten (ch) << 8) | (WasTouched (ch) << 4) | (WasLatched (ch))
    );

  ldebug ("base %llu counter %llu touched at %llu\n\n",
          ch->base, ch->counter, ch->touched_at);
}

static inline int flipflop (struct ctc_chan *ch)
{
  int ff;

  ff = ch->status & 1;
  ch->status ^= 1;
  return ff;
}

static inline ul calc_elapsed (ul base)
{
  ul nsec;

  nsec = TID_get_nsec_diff (base);
  return nsec;
}

static us latch (struct ctc_chan *ch)
{
  ul elapsed, counter, reload;
  ui ticks;

  if ((3 == Mode (ch)) && mode3_default_latch) {
    return mode3_default_latch;
  }

  if (WasTouched (ch)) {
    elapsed = calc_elapsed (ch->touched_at);
    ClrTouched (ch);
  }
  else {
    elapsed = calc_elapsed (ch->base);
  }

  counter = ch->counter;
  reload = (0 == ch->reload) ? 0x10000 : ch->reload;
  ticks = (reload * elapsed) / counter;

  log_chan (__func__, ch);

  if (ticks > reload) {
    ticks = (ticks - reload) % reload;
  }

  if (3 == Mode (ch)) {
    int result;

    if (ticks > reload / 2) {
      result = reload - ticks;
    }
    else {
      result = reload - (ticks * 2);
    }
    return result;
  }
  else {
    return reload - ticks;
  }
}

static ui read_40 (int nport, int size)
{
  ui ticks, retval;
  struct ctc_chan *ch;

  ch = channels;

  DEREF (nport);
  DEREF (size);

  EIO_LOG_READ;

  log_chan (__func__, ch);

  if (WasLatched (ch)) {
    ticks = ch->latch;
  }
  else {
    lwarn ("read without latching\n");
#if 1
    ticks = latch (ch);
#else
    ch->latch = latch (ch);
    ticks = ch->latch;
    SetLatched (ch);
#endif
  }

  switch (AccessMode (ch)) {
    case AccessLoByte:
      retval = ticks;
      break;

    case AccessHiByte:
      retval = ticks >> 8;
      break;

    case AccessLoHi:
      retval = ticks >> (flipflop (ch) << 3);
      break;

    default:
      Aldri ();
  }

  return retval;
}

static ui read_nothing (int nport, int size)
{
  DEREF (nport);
  DEREF (size);

  lwarn ("%#x %d\n", nport, size);
  return ~0U;
}

static inline ul reload_to_counter (us reload)
{
  ul counter;

  counter = (0 == reload) ? 0x10000 : reload;
  counter = (TID_NSEC * counter) / PIT_HZ;
  return counter;
}

static void write_40 (int nport, int size, ui val)
{
  int reload;
  struct ctc_chan *ch;

  DEREF (nport);
  DEREF (size);

  EIO_LOG_WRITE;

  ch = channels;

  val &= 0xff;

  reload = ch->new_reload;
  switch (AccessMode (ch)) {
    case AccessLoByte:
      reload = val;
      break;

    case AccessHiByte:
      reload = val << 8;
      break;

    case AccessLoHi:
      reload |= val << (flipflop (ch) << 3);
      break;

    default:
      Aldri ();
  }

  if (0 == Mode (ch))
    ClrOutPin (ch);

  if ((AccessMode (ch) != AccessLoHi) || (0 == GetFlipFlop (ch))) {
    SetTouched (ch);
    ch->touched_at = TID_get_abs_nsec ();

    if ((reload != ch->reload) || (0 == ch->base)) {
      ch->reload = reload;
      SetWritten (ch);
    }
    if (0 == ch->base) {
      ch->base = ch->touched_at;
    }
  }
  log_chan (__func__, ch);
}

static void write_43 (int nport, int size, ui val)
{
  int channel;
  int access_mode;
  int mode;
  int bcd;
  char *msg;
  struct ctc_chan *ch;

  DEREF (nport);
  DEREF (size);

  EIO_LOG_WRITE;

  msg = (char *) -1;
  channel = (val >> 6) & 3;
  access_mode = (val >> 4) & 3;
  mode = (val >> 1) & 7;
  bcd = val & 1;

  if (1 == bcd) {
    msg = "BCD not supported";
    goto error;
  }

  if (0 != channel) {
    /* msg = "only channel 0 is supported"; */
    /* goto error; */
    lwarn ("channel %d is not supported\n", channel);
    return;
  }

  if ((0 != access_mode) && ((mode > 3) || (1 == mode))) {
    msg = "only modes 0, 2, 3 are supported";
    goto error;
  }

  ch = channels + channel;

  if (access_mode) {
    ch->val = val;
    ch->new_reload = 0;
    ch->status = 0;

    switch (mode) {
      case 0:
        ClrOutPin (ch);
        break;

      case 2:
      case 3:
        SetOutPin (ch);
        break;
    }

    log_chan (__func__, ch);
  }
  else {
    ch->latch = latch (ch);
    ClrFlipFlop (ch);
    SetLatched (ch);
  }

  return;

 error:
  Fail ("val         = %#x %s\n"
        "channel     = %d\n"
        "access_mode = %d %s\n"
        "mode        = %d\n"
        "bcd         = %d\n",
        val, msg,
        channel,
        access_mode, access_names[access_mode],
        mode,
        bcd);
}

void PIT_run ()
{
  struct ctc_chan *ch;
  ul elapsed, counter;
  static int in;

  Fail_if (in);

  in = 1;
  ch = channels;

  elapsed = calc_elapsed (ch->base);
  counter = ch->counter;

  if (elapsed >= counter) {

    if (WasWritten (ch)) {
      ch->base = ch->touched_at;
      ch->counter = reload_to_counter (ch->reload);
      ClrWritten (ch);
      ClrTouched (ch);
    }
    else {
      ch->base += counter;
    }

    switch (Mode (ch)) {
      case 0: {
        if (0 == GetOutPin (ch)) {
          SetOutPin (ch);
          PIC_request_service (0);
        }
        break;
      }

      case 3:
      case 2:
        PIC_request_service (0);
        break;
    }
  }

  in = 0;
}

static void IRQ0 (INT_ARGS)
{
  DEREF (r);

  MEM_TICKS += 1;
  PIC_eoi (0);
}

static void pic_handler (int which)
{
  if (which) {
    PIT_run ();
    linfo ("finished running timer, pending %d\n", PIC_irq_pending ());
  }
  else {
    linfo ("running timer\n");
  }
}

void PIT_pre_init ()
{
  static struct MUP_key keys[] = {
    {"pit.mode3.latch"      , TAG_IN_INT , {.ii = 17050}},
  };

  MUP_add_keys (keys, LENOFA (keys));
}

void PIT_init ()
{
  log_level = MUP_get_log_level ("pit");

  mode3_default_latch = MUP_get_int ("pit.mode3.latch");

  EIO_register_write_handler (0x40, EIO_MASK_B, write_40);
  EIO_register_write_handler (0x43, EIO_MASK_B, write_43);
  EIO_register_read_handler (0x40, EIO_MASK_B, read_40);

  EIO_register_read_handler (0x41, EIO_MASK_B, read_nothing);
  EIO_register_read_handler (0x42, EIO_MASK_B, read_nothing);
  INT_set (8, IRQ0);

  PIC_register (0, pic_handler);

  write_43 (43, 1, (3 << 4) | (2 << 1));
  write_40 (40, 1, 00);
  write_40 (40, 1, 00);
  channels->counter = reload_to_counter (channels->reload);
}
