#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <i86/turtle.h>

#undef SET_CF
#include <inc/x86.h>
#include <inc/asm.h>
#include <inc/int.h>
#include <inc/mem.h>
#include <inc/ldt.h>
#include <inc/xdt.h>
#include <inc/sig.h>
#include <inc/sel.h>
#include <inc/eio.h>
#include <inc/dpmi.h>
#include <inc/mup.h>
#include <inc/log.h>

#include <vgaemu/misc.h>
#include <vgaemu/config.h>
#include <vgaemu/vgaemu.h>
#include <vgaemu/cpu-all.h>

static char *turtle_log_name;
extern FILE *turtle_log;
FILE *turtle_log;

static const char * const reg_names[8] = {
  "eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi"
};

static const char * const seg_names[6] = {
  "es", "cs", "ss", "ds", "fs", "gs"
};

static CPUReadMemoryFunc **vga_read_funcs;
static CPUWriteMemoryFunc **vga_write_funcs;

static CPUReadMemoryFunc *turtle_read_vga_mem8;
static CPUReadMemoryFunc *turtle_read_vga_mem16;
static CPUReadMemoryFunc *turtle_read_vga_mem32;

static CPUWriteMemoryFunc *turtle_write_vga_mem8;
static CPUWriteMemoryFunc *turtle_write_vga_mem16;
static CPUWriteMemoryFunc *turtle_write_vga_mem32;

static int log_level;
#define LOG_PREFIX "X86"

#define TOEMU(env) ((struct X86_emu*)(env))
#define HOT(env) ((struct X86_emu*)(env))->hot=1
#define IS_MEM_VGA(env,a) (((a)>=TOEMU(env)->mem_l)&&((a)<TOEMU(env)->mem_u))

struct X86_emu {
  X86 env;
  ui mem_l, mem_u;
  int hot;
  int exit;
  int threshold;
  int replay;
};

static int catch_VGA;
static struct X86_emu *current;
static struct X86 before_env, after_env;

const uint8_t parity_tab[256] =
  { PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF,
    0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0, PF, 0,
    PF, 0, PF, 0, };

char *X86_efltoa (ui eflags, char m[31])
{
  char *s;

  s = m;
#define C(m,c) *s++ = c ^ ((!!(eflags & m)) << 5)
#define F(n,c) C(n##F,c)
  F(C,'c'); F(P,'p'); F(A,'a'); F(Z,'z'); F(S,'s');
  F(T,'t'); F(I,'i'); F(D,'d'); F(O,'o'); C(VM,'v');
#undef C
#undef F
  snprintf (s, 15,
            " IOPL=%d%s%s%s%s%s", /* 6+1+3*3+2*4) = 24 */
            (eflags >> IOPL_SHIFT) & 3,
            (eflags & NT ? " NT" : ""),
            (eflags & RF ? " RF" : ""),
            (eflags & AC ? " AC" : ""),
            (eflags & VIF ? " VIF" : ""),
            (eflags & VIP ? " VIP" : ""));
  /* 25 */
  /* *s = 0; */
  return m;
}

const char *X86_reg_name (ub nreg)
{
  Fail_if (nreg > 7);
  return reg_names[nreg];
}

const char *X86_seg_name (ub nseg)
{
  Fail_if (nseg > 5);
  return seg_names[nseg];
}

int X86_process_prefixes (ub *pc, ui *ppref, int *poverride)
{
  int len, override;
  ui prefix;

  prefix = 0;
  override = -1;

  for (len = 0; ; len++) {
    switch (pc[len]) {
      case 0x26:
        override = X86_PREFIX_ES;
        continue;

      case 0x2e:
        override = X86_PREFIX_CS;
        continue;

      case 0x36:
        override = X86_PREFIX_SS;
        continue;

      case 0x3e:
        override = X86_PREFIX_DS;
        continue;

      case 0x64:
        override = X86_PREFIX_FS;
        continue;

      case 0x65:
        override = X86_PREFIX_GS;
        continue;

      case 0x66:
        prefix |= X86_PREFIX_DATA;
        continue;

      case 0x67:
        prefix |= X86_PREFIX_ADDR;
        continue;

      case 0xf0:
        prefix |= X86_PREFIX_LOCK;
        continue;

      case 0xf2:
        prefix |= X86_PREFIX_REPNE;
        continue;

      case 0xf3:
        prefix |= X86_PREFIX_REP;
        continue;
    }
    break;
  }

  if (ppref) *ppref = prefix;
  if (poverride) *poverride = override;
  return len;
}

static void log_emu (struct X86_emu *emu)
{
  X86 *env;
  char m[X86_EFLAGS_BUF_SIZE];

  env = &emu->env;

  log ("\neax=%#010x ebx=%#010x ecx=%#010x edx=%#010x\n"
        "esi=%#010x edi=%#010x ebp=%#010x esp=%#010x\n"
        "eip=%#010x efl=%#010x %s\n"
        "cs=%#06x ds=%#06x es=%#06x ss=%#06x "
        "fs=%#06x gs=%#06x\n\n",
        env->regs[_REG_EAX].d, env->regs[_REG_EBX].d,
        env->regs[_REG_ECX].d, env->regs[_REG_EDX].d,
        env->regs[_REG_ESI].d, env->regs[_REG_EDI].d,
        env->regs[_REG_EBP].d, env->regs[_REG_ESP].d,
        env->eip, env->eflags,
        X86_efltoa (env->eflags, m),
        env->segs[_SEG_CS].selector,
        env->segs[_SEG_DS].selector,
        env->segs[_SEG_ES].selector,
        env->segs[_SEG_SS].selector,
        env->segs[_SEG_FS].selector,
        env->segs[_SEG_GS].selector
    );
}

uint32_t P__(lea, uint8_t modrm)
{
  int mod;
  int rm;

  mod = (modrm >> 6) & 3;
  rm = modrm & 7;

  /* assert (3 != mod); */

  if (env->a32) {
    uint32_t disp;

    disp = 0;
    if (4 == rm) {
      int index, scale;
      uint8_t sib;

      sib = FETCH_BYTE (env);

      rm = sib & 7;
      index = (sib >> 3) & 7;
      scale = (sib >> 6) & 3;

      if (4 != index)
        disp += READ_DWORD_REG (env, index) << scale;
    }

    if (1 == mod) disp += (int8_t) FETCH_BYTE (env);
    disp += READ_DWORD_REG (env, rm);
    if ((rm | -mod) == 5) {
      disp += FETCH_DWORD (env);
      disp -= READ_DWORD_REG (env, 5);
    }
    else {
      if ((0 == (env->iflags & IF_SEG_OVERRIDE)) && (4 == (rm & 6)))
        env->seg = _SEG_SS;

      if (2 == mod) disp += FETCH_DWORD (env);
    }

    return disp;
  }
  else {
    uint32_t disp = 0;
    switch (mod) {
      case 0:
        if (6 == rm)
          return FETCH_WORD (env);
        break;

      case 2:
        disp = FETCH_WORD (env);
        break;

      case 1:
        disp = (int8_t) FETCH_BYTE (env);
        break;
    }

    if (rm < 6)
      disp += READ_WORD_REG (env, _REG_SI + (rm & 1));

    switch (rm) {
      case 0:                 /* 000 */
      case 1:                 /* 001 */
      case 7:                 /* 111 */
        disp += READ_WORD_REG (env, _REG_BX);
        break;

      case 2:                 /* 010 */
      case 3:                 /* 011 */
      case 6:                 /* 110 */
        disp += READ_WORD_REG (env, _REG_BP);
        if (0 == (env->iflags & IF_SEG_OVERRIDE))
          env->seg = _SEG_SS;
        break;
    }
    return (disp & 0xffff);
  }
}

void P__(write_cr, int rno, uint32_t v)
{
  DEREF (v);
  DEREF (rno);
  turtle_not_implemented (env);
}

void P__(write_dr, int rno, uint32_t v)
{
  DEREF (v);
  DEREF (rno);
  turtle_not_implemented (env);
}

void P0__(not_implemented)
{
  env->iflags |= IF_NOT_IMPLEMENTED;
}

void P0__(confirm)
{
  env->iflags |= IF_CONFIRM;
  memcpy (&before_env, env, sizeof (before_env));
}

void P__(exception, int intno, int err)
{
  if (LOG_DEBUG == log_level) {
    ldebug ("exception %d error %d\n", intno, err);
    log_emu (TOEMU (env));
  }
  env->iflags |= IF_NOT_IMPLEMENTED;
  Fail ("EXCEPTION %#x error %#x", intno, err);
}

void P__(interrupt, int intno)
{
  DEREF (intno);

  if (LOG_DEBUG == log_level) {
    ldebug ("interrupt %#x\n", intno);
    log_emu (TOEMU (env));
  }
  env->iflags |= IF_NOT_IMPLEMENTED;
}

static x86exit_e interp (struct X86_emu *emu)
{
  int i, insns;
  X86 *env;

  env = &emu->env;
  i = emu->threshold;

  insns = 0;
  while (i--) {
    int intr;

    if (emu->exit && (0 == (env->iflags & IF_INTR_DISABLED))) {
      ldebug ("exit requested, processed total of %d insns\n", insns);
      return X86_EXIT_REQUESTED;
    }

    env->pc = env->segs[_SEG_CS].base + env->eip;
    env->rep = 0;
    env->a32 = env->o32 = env->c32;
    env->seg = _SEG_DS;

    intr = env->iflags & IF_INHIBIT_INTS;

    env->iflags |= IF_INTR_DISABLED;
    env->iflags &= ~(IF_INHIBIT_INTS | IF_JUMP | IF_SEG_OVERRIDE);
    turtle (env);

    if (env->iflags & IF_NOT_IMPLEMENTED) {
      if (env->segs[_SEG_CS].base[env->eip] != 0xcd) {
        ldebug ("bad at %p %#x %#x %#x\n",
                env->segs[_SEG_CS].base + env->eip,
                env->segs[_SEG_CS].base[env->eip],
                env->segs[_SEG_CS].base[env->eip + 1],
                env->segs[_SEG_CS].base[env->eip + 2]);
      }

      ldebug ("processed total of %d insns\n", insns);
      return X86_EXIT_BAD_INSN;
    }

    if (env->iflags & IF_CONFIRM) {
      ldebug ("confirmation requested, processed total of %d insns\n", insns);
      return X86_EXIT_TEMPORARY;
    }

    if (0 == (env->iflags & IF_JUMP)) {
      env->eip = NEXT_IP (env);
    }

    insns++;

    env->iflags ^= intr;

    if (0 == (env->iflags & IF_INHIBIT_INTS)) {
      env->iflags &= ~IF_INTR_DISABLED;
    }

    if (emu->hot) {
      emu->hot = 0;
      i = emu->threshold;
    }
  }

  ldebug ("threshold processed total of %d insns\n", insns);
  return X86_EXIT_THRESHOLD;
}

uint8_t P__(read_port8, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 1);
}

uint16_t P__(read_port16, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 2);
}

uint32_t P__(read_port32, uint16_t nport)
{
  HOT (env);
  return EIO_read (nport, 4);
}

void P__(write_port8, uint16_t nport, uint8_t v)
{
  HOT (env);
  EIO_write (nport, 1, v);
}

void P__(write_port16, uint16_t nport, uint16_t v)
{
  HOT (env);
  EIO_write (nport, 2, v);
}

void P__(write_port32, uint16_t nport, uint32_t v)
{
  HOT (env);
  EIO_write (nport, 4, v);
}

#define SIZE 8
#include <inc/mem.inc>
#undef SIZE

#define SIZE 16
#include <inc/mem.inc>
#undef SIZE

#define SIZE 32
#include <inc/mem.inc>
#undef SIZE

/* pushes */
void P__(push16, uint16_t v)
{
  if (IS_STACK_32 (env)) {
    env->regs[_REG_ESP].d -= 2;
    if (0 == TOEMU (env)->replay)
      *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d) = v;
  }
  else {
    env->regs[_REG_SP].w -= 2;
    if (0 == TOEMU (env)->replay)
      *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w) = v;
  }
}

void P__(push32, uint32_t v)
{
  if (IS_STACK_32 (env)) {
    env->regs[_REG_ESP].d -= 4;
    if (0 == TOEMU (env)->replay)
      *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d) = v;
  }
  else {
    env->regs[_REG_SP].w -= 4;
    if (0 == TOEMU (env)->replay)
      *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w) = v;
  }
}

uint16_t P0__(pop16)
{
  uint16_t v;

  v = 0;
  if (IS_STACK_32 (env)) {
    if (0 == TOEMU (env)->replay)
      v = *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d);
    env->regs[_REG_ESP].d += 2;
  }
  else {
    if (0 == TOEMU (env)->replay)
      v = *(uint16_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w);
    env->regs[_REG_SP].w += 2;
  }

  return v;
}

uint32_t P0__(pop32)
{
  uint32_t v;

  v = 0;
  if (IS_STACK_32 (env)) {
    if (0 == TOEMU (env)->replay)
      v = *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_ESP].d);
    env->regs[_REG_ESP].d += 4;
  }
  else {
    if (0 == TOEMU (env)->replay)
      v = *(uint32_t *) (env->segs[_SEG_SS].base + env->regs[_REG_SP].w);
    env->regs[_REG_SP].w += 4;
  }
  return v;
}

void P0__(cli)
{
#ifdef IF_MANAGEMENT
  uint32_t f;

  f = env->eflags;

  if (0 == PE (env)) {
    env->eflags = f & ~ IF;
  }
  else {
    int cpl, iopl;

    cpl = CPL (env);
    iopl = (f & IOPL_MASK) >> IOPL_SHIFT;

    if (0 == (f & VM)) {
      if (iopl >= cpl) {
        env->eflags = f & ~ IF;
      }
      else {
        if ((iopl < cpl) && (cpl < 3) && PVI (env)) {
          env->eflags = f & ~ VIF;
        }
        else {
          turtle_exception (env, 13, 0);
        }
      }
    }
    else {
      if (3 == iopl) {
        env->eflags = f & ~ IF;
      }
      else {
        if ((iopl < 3) && VME (env)) {
          env->eflags = f & ~ VIF;
        }
        else {
          turtle_exception (env, 13, 0);
        }
      }
    }
  }
#else
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(sti)
{
#ifdef IF_MANAGEMENT
  uint32_t f;

  f = env->eflags;

  if (0 == PE (env)) {
    env->eflags = f | IF;
  }
  else {
    int cpl;
    int iopl;

    cpl = CPL (env);
    iopl = (f & IOPL_MASK) >> IOPL_SHIFT;

    if (0 == (f & VM)) {
      if (iopl >= cpl) {
        env->eflags = f | IF;
        INHIBIT_INTERRUPTS (env);
      }
      else {
        if ((iopl < cpl) && (3 == cpl) && (0 == (f & VIP))) {
          env->eflags = f | VIF;
        }
        else {
          env->eflags = f;
          EXCEPTION (env, 13, 0);
        }
      }
    }
    else {
      if (3 == iopl) {
        env->eflags = f | IF;
        INHIBIT_INTERRUPTS (env);
      }
      else {
        if ((iopl < 3) && (0 == (f & VIP)) && VME (env)) {
          env->eflags = f | VIF;
        }
        else {
          EXCEPTION (env, 13, 0);
        }
      }
    }
  }
#else
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

static void pushf (X86 *env, int d)
{
#ifdef IF_MANAGEMENT
  int pe, vm, iopl;

  pe = PE (env);
  vm = 0 != (env->eflags & VM);
  iopl = (env->eflags & IOPL_MASK) >> IOPL_SHIFT;

  if ((0 == pe) || (pe && ((0 == vm) || (vm && (3 == iopl))))) {
    if (d)
      turtle_push32 (env, env->eflags & 0xfcffff);
    else
      turtle_push16 (env, env->eflags);
  }
  else {
    turtle_exception (env, 13, 0);
  }
#else
  DEREF (env);
  DEREF (d);
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(pushf16)
{
  pushf (env, 0);
}

void P0__(pushf32)
{
  pushf (env, 1);
}

static void popf (X86 *env, int d)
{
#ifdef IF_MANAGEMENT
  uint32_t rF, wF;

  rF = env->eflags;
  wF = rF;
  if (0 == (rF & VM)) {
    int cpl;

    cpl = CPL (env);
    if (d)
      wF = turtle_pop32 (env);
    else
      wF = turtle_pop16 (env);

    SANITIZE_EFLAGS (wF);

    if (0 == cpl) {
      wF &= ~ (VIF | VIP | VM);
      wF |= rF & VM;
    }
    else {
      wF &= ~ (VIF | VIP | VM | IOPL_MASK);
      wF |= rF & (VM | (3 << IOPL_SHIFT));
    }
  }
  else {
    if (3 == ((rF & IOPL_MASK) >> IOPL_SHIFT)) {
      if (d)
        wF = turtle_pop32 (env);
      else
        wF = turtle_pop16 (env);
      SANITIZE_EFLAGS (wF);
      wF &= ~ (VIF | VIP | VM | RF | IOPL_MASK);
      wF |= rF & (VIF | VIP | VM | RF | IOPL_MASK);
    }
    else {
      EXCEPTION (env, 13, 0);
    }
  }
  if ((rF & IF) && (0 == (wF & IF))) {
    ldebug ("popf disabled interrupts\n");
  }
  env->eflags = wF;
#else
  DEREF (env);
  DEREF (d);
  env->iflags |= IF_NOT_IMPLEMENTED;
#endif
}

void P0__(popf16)
{
  popf (env, 0);
}

void P0__(popf32)
{
  popf (env, 1);
}

void P__(load_segment, int seg, uint32_t selector)
{
  if (PE (env) && (0 == (env->eflags & VM))) {
    unsigned int ar;
    int limit;
    unsigned long base;
    /* printf ("loading seg %#x selector %#x\n", seg, selector); */

    if (0x40 == selector)
      selector = DPMI_get_bios_selector ();

    if (IS_GDT_SELECTOR (selector)) {
      if (ASM_cs == selector)
        ar = 0xcffa00;
      else
        ar = 0xcff300;
      limit = -1;
      base = 0;
    }
    else {
      ul d;

      LDT_get_selector_descriptor (selector, &d);
      limit = XDT_get_descriptor_limit (&d);
      base = XDT_get_descriptor_base (&d);
      ar = (uint32_t) (d >> 32U);
    }
    env->segs[seg].selector = (selector & 0xffff);
    env->segs[seg].base = (uint8_t *) base;
    env->segs[seg].limit = limit;
    env->segs[seg].flags = ar;

    if (_SEG_CS == seg) env->c32 = (ar >> 22) & 1;
    if (_SEG_SS == seg) env->s32 = (ar >> 22) & 1;
  }
  else {
    env->segs[seg].selector = (selector & 0xffff);
    env->segs[seg].base = (uint8_t *) ((selector & 0xffff) << 4);
    env->segs[seg].limit = 0xffff;
    env->segs[seg].flags = 0;
    if (_SEG_CS == seg) env->c32 = 0;
    if (_SEG_SS == seg) env->s32 = 0;
  }
}

static uint32_t read_nothing (uint32_t addr)
{
  DEREF (addr);
  return 0;
}

static void write_nothing (uint32_t addr, uint32_t value)
{
  DEREF (addr);
  DEREF (value);
}

int X86_replay_stack (regs86 *r, ui stop_eip)
{
  FILE *save;
  X86 *env;
  int insns, retval;
  ui flags_esp;
  struct X86_emu emu;

  memset (&emu, 0, sizeof (emu));

  save = turtle_log;
  turtle_log = NULL;

  env = &emu.env;
  env->cr[0] = 1;
  env->eip = SIGEIP (r);
  env->iflags = 0;

  emu.hot = 0;
  emu.exit = 0;
  emu.mem_l = 0;
  emu.mem_u = ~0U;
  emu.replay = 1;

  WRITE_DWORD_REG (env, _REG_EAX, SIGEAX (r));
  WRITE_DWORD_REG (env, _REG_EBX, SIGEBX (r));
  WRITE_DWORD_REG (env, _REG_ECX, SIGECX (r));
  WRITE_DWORD_REG (env, _REG_EDX, SIGEDX (r));
  WRITE_DWORD_REG (env, _REG_ESI, SIGESI (r));
  WRITE_DWORD_REG (env, _REG_EDI, SIGEDI (r));
  WRITE_DWORD_REG (env, _REG_EBP, SIGEBP (r));
  WRITE_DWORD_REG (env, _REG_ESP, SIGESP (r));
  WRITE_SEGMENT_REG (env, _SEG_CS, SIGCS (r));
  WRITE_SEGMENT_REG (env, _SEG_DS, SIGDS (r));
  WRITE_SEGMENT_REG (env, _SEG_ES, SIGES (r));
  WRITE_SEGMENT_REG (env, _SEG_SS, SIGSS (r));
  WRITE_SEGMENT_REG (env, _SEG_FS, SIGFS (r));
  WRITE_SEGMENT_REG (env, _SEG_GS, SIGGS (r));

  turtle_write_vga_mem8 = write_nothing;
  turtle_write_vga_mem16 = write_nothing;
  turtle_write_vga_mem32 = write_nothing;

  turtle_read_vga_mem8 = read_nothing;
  turtle_read_vga_mem16 = read_nothing;
  turtle_read_vga_mem32 = read_nothing;

  flags_esp = SIGESP (r);
  retval = -1;

  for (insns = 0; env->eip != stop_eip; insns++) {
    env->pc = env->segs[_SEG_CS].base + env->eip;
    env->rep = 0;
    env->a32 = env->o32 = env->c32;
    env->seg = _SEG_DS;

    env->iflags &= ~(IF_INHIBIT_INTS | IF_JUMP | IF_SEG_OVERRIDE);
    turtle (env);

    if (READ_DWORD_REG (env, _REG_ESP) > flags_esp) {
      int len;
      uint8_t *pc, opc;

      pc = env->segs[_SEG_CS].base + env->eip;
      len = X86_process_prefixes (pc, NULL, NULL);
      opc = pc[len];

      if ((opc >= 0x58) && (opc < 0x60)) {
        retval = opc & 7;
        goto exit;
      }
      else
        goto exit;
    }

    if (env->iflags & IF_NOT_IMPLEMENTED) {
      if (env->segs[_SEG_CS].base[env->eip] != 0xcd) {
        ldebug ("bad at %p %#x %#x %#x\n",
                env->segs[_SEG_CS].base + env->eip,
                env->segs[_SEG_CS].base[env->eip],
                env->segs[_SEG_CS].base[env->eip + 1],
                env->segs[_SEG_CS].base[env->eip + 2]);
      }

      ldebug ("processed total of %d insns\n", insns);
      goto exit;
    }

    if (0 == (env->iflags & IF_JUMP)) {
      env->eip = NEXT_IP (env);
      if ((env->eip > stop_eip) || (env->eip < (uint32_t) SIGEIP (r)))
        goto exit;
    }
  }

  SIGEAX (r) = READ_DWORD_REG (env, _REG_EAX);
  SIGEBX (r) = READ_DWORD_REG (env, _REG_EBX);
  SIGECX (r) = READ_DWORD_REG (env, _REG_ECX);
  SIGEDX (r) = READ_DWORD_REG (env, _REG_EDX);
  SIGESI (r) = READ_DWORD_REG (env, _REG_ESI);
  SIGEDI (r) = READ_DWORD_REG (env, _REG_EDI);
  SIGEBP (r) = READ_DWORD_REG (env, _REG_EBP);
  SIGESP (r) = READ_DWORD_REG (env, _REG_ESP);
  SIGCS (r) = READ_SEGMENT_REG (env, _SEG_CS);
  SIGDS (r) = READ_SEGMENT_REG (env, _SEG_DS);
  SIGES (r) = READ_SEGMENT_REG (env, _SEG_ES);
  SIGSS (r) = READ_SEGMENT_REG (env, _SEG_SS);
  SIGFS (r) = READ_SEGMENT_REG (env, _SEG_FS);
  SIGGS (r) = READ_SEGMENT_REG (env, _SEG_GS);
  SIGEIP (r) = env->eip;
  SIGEFL (r) = env->eflags;

  retval = 8;

 exit:
  if (vga_write_funcs) {
    turtle_write_vga_mem8 = vga_write_funcs[0];
    turtle_write_vga_mem16 = vga_write_funcs[1];
    turtle_write_vga_mem32 = vga_write_funcs[2];

    turtle_read_vga_mem8 = vga_read_funcs[0];
    turtle_read_vga_mem16 = vga_read_funcs[1];
    turtle_read_vga_mem32 = vga_read_funcs[2];
  }

  turtle_log = save;
  return retval;
}

x86exit_e X86_confirm32 (regs86 *r, int threshold)
{
  int i;
  ui mask;
  X86 *env;
  int diff;

  env = &after_env;
  diff = 0;

  for (i = 0; i < 8; i++) {
    uint32_t hw, sw;

    hw = *SIG_gregp (r, i);
    sw = READ_DWORD_REG (env, i);
    if (hw != sw) {
      log ("%s %#010x %#010x\n", reg_names[i], hw, sw);
      diff = 1;
    }
  }

  for (i = 0; i < 6; i++) {
    uint16_t hw, sw;

    hw = *SIG_sregp (r, i);
    sw = READ_SEGMENT_REG (env, i);
    if (hw != sw) {
      log ("%s %#06x %#06x\n", seg_names[i], hw, sw);
    }
    diff = 1;
  }

  if ((uint32_t) SIGEIP (r) != env->eip) {
    log ("EIP %p %p\n",
         env->segs[_SEG_CS].base + SIGEIP (r),
         env->segs[_SEG_CS].base + env->eip);
    diff = 1;
  }

  mask = IF|TF|RF|(3<<IOPL_SHIFT);
  if ((SIGEFL (r) & ~mask) != (env->eflags & ~mask)) {
    char m1[X86_EFLAGS_BUF_SIZE];
    char m2[X86_EFLAGS_BUF_SIZE];

    X86_efltoa (SIGEFL (r), m1);
    X86_efltoa (env->eflags, m2);

    log ("%#10x EFL %#7x %s %#7x %s\n",
         (int) SIGEIP (r), (int) SIGEFL (r), m1, env->eflags, m2);
    diff = 0;
  }
  if (diff) {
    log ("DIFFERENT\n");
    abort ();
  }
  return X86_run32 (r, threshold);
}

x86exit_e X86_run32 (regs86 *r, int threshold)
{
  FILE *save;
  X86 *env;
  int was_enabled;
  x86exit_e exit;
  struct X86_emu emu;

  memset (&emu, 0, sizeof (emu));

  save = turtle_log;
  turtle_log = NULL;

  env = &emu.env;
  env->cr[0] = 1;

  was_enabled = INT_is_enabled ();
  env->eflags = SIGEFL (r);

  if (was_enabled)
    env->eflags |= IF_MASK;
  else
    env->eflags &= ~IF_MASK;

  env->eip = SIGEIP (r);
  env->iflags = 0;

  WRITE_DWORD_REG (env, _REG_EAX, SIGEAX (r));
  WRITE_DWORD_REG (env, _REG_EBX, SIGEBX (r));
  WRITE_DWORD_REG (env, _REG_ECX, SIGECX (r));
  WRITE_DWORD_REG (env, _REG_EDX, SIGEDX (r));
  WRITE_DWORD_REG (env, _REG_ESI, SIGESI (r));
  WRITE_DWORD_REG (env, _REG_EDI, SIGEDI (r));
  WRITE_DWORD_REG (env, _REG_EBP, SIGEBP (r));
  WRITE_DWORD_REG (env, _REG_ESP, SIGESP (r));
  WRITE_SEGMENT_REG (env, _SEG_CS, SIGCS (r));
  WRITE_SEGMENT_REG (env, _SEG_DS, SIGDS (r));
  WRITE_SEGMENT_REG (env, _SEG_ES, SIGES (r));
  WRITE_SEGMENT_REG (env, _SEG_SS, SIGSS (r));
  WRITE_SEGMENT_REG (env, _SEG_FS, SIGFS (r));
  WRITE_SEGMENT_REG (env, _SEG_GS, SIGGS (r));

  turtle_log = save;

  if (0x806a064 == SIGEIP (r))
    SIG_break (0);

  emu.threshold = threshold;
  emu.hot = 0;
  emu.exit = 0;
  emu.replay = 0;

  if (catch_VGA) {
    emu.mem_l = MEM_REAL (0xa0000);
    emu.mem_u = MEM_REAL (0xc0000);
  }

  ldebug ("interpreting at %p\n", env->segs[_SEG_CS].base + env->eip);

  current = &emu;
  exit = interp (&emu);
  current = NULL;

  if (X86_EXIT_TEMPORARY == exit) {
    memcpy (&after_env, env, sizeof (after_env));
    after_env.eip = NEXT_IP ((&after_env));
    env = &before_env;
  }

  SIGEAX (r) = READ_DWORD_REG (env, _REG_EAX);
  SIGEBX (r) = READ_DWORD_REG (env, _REG_EBX);
  SIGECX (r) = READ_DWORD_REG (env, _REG_ECX);
  SIGEDX (r) = READ_DWORD_REG (env, _REG_EDX);
  SIGESI (r) = READ_DWORD_REG (env, _REG_ESI);
  SIGEDI (r) = READ_DWORD_REG (env, _REG_EDI);
  SIGEBP (r) = READ_DWORD_REG (env, _REG_EBP);
  SIGESP (r) = READ_DWORD_REG (env, _REG_ESP);
  SIGCS (r) = READ_SEGMENT_REG (env, _SEG_CS);
  SIGDS (r) = READ_SEGMENT_REG (env, _SEG_DS);
  SIGES (r) = READ_SEGMENT_REG (env, _SEG_ES);
  SIGSS (r) = READ_SEGMENT_REG (env, _SEG_SS);
  SIGFS (r) = READ_SEGMENT_REG (env, _SEG_FS);
  SIGGS (r) = READ_SEGMENT_REG (env, _SEG_GS);
  SIGEIP (r) = env->eip;
  SIGEFL (r) = env->eflags;

  if (env->eflags & IF) {
    INT_enable ();
  }
  else {
    INT_disable ();
  }

  turtle_log = save;
  return exit;
}

x86exit_e X86_run16 (struct vm86_regs *vm86, int threshold)
{
  FILE *save;
  X86 *env;
  x86exit_e exit;
  struct X86_emu emu;

  memset (&emu, 0, sizeof (emu));

  save = turtle_log;
  turtle_log = NULL;

  env = &emu.env;
  env->cr[0] = 1;
  env->eflags = vm86->eflags | VM | (3 << IOPL_SHIFT);
  env->eip = vm86->eip;
  env->iflags = 0;

  WRITE_DWORD_REG (env, _REG_EAX, vm86->eax);
  WRITE_DWORD_REG (env, _REG_EBX, vm86->ebx);
  WRITE_DWORD_REG (env, _REG_ECX, vm86->ecx);
  WRITE_DWORD_REG (env, _REG_EDX, vm86->edx);
  WRITE_DWORD_REG (env, _REG_ESI, vm86->esi);
  WRITE_DWORD_REG (env, _REG_EDI, vm86->edi);
  WRITE_DWORD_REG (env, _REG_EBP, vm86->ebp);
  WRITE_DWORD_REG (env, _REG_ESP, vm86->esp);
  WRITE_SEGMENT_REG (env, _SEG_CS, vm86->cs);
  WRITE_SEGMENT_REG (env, _SEG_DS, vm86->ds);
  WRITE_SEGMENT_REG (env, _SEG_ES, vm86->es);
  WRITE_SEGMENT_REG (env, _SEG_SS, vm86->ss);
  WRITE_SEGMENT_REG (env, _SEG_FS, vm86->fs);
  WRITE_SEGMENT_REG (env, _SEG_GS, vm86->gs);

  emu.threshold = threshold;
  emu.hot = 0;
  emu.exit = 0;
  emu.replay = 0;
  if (catch_VGA) {
    emu.mem_l = MEM_REAL (0xa0000);
    emu.mem_u = MEM_REAL (0xc0000);
  }

  ldebug ("interpreting at %p\n", env->segs[_SEG_CS].base + env->eip);
  current = &emu;
  exit = interp (&emu);
  /* log_emu (&emu); */
  current = NULL;

  vm86->eax = READ_DWORD_REG (env, _REG_EAX);
  vm86->ebx = READ_DWORD_REG (env, _REG_EBX);
  vm86->ecx = READ_DWORD_REG (env, _REG_ECX);
  vm86->edx = READ_DWORD_REG (env, _REG_EDX);
  vm86->esi = READ_DWORD_REG (env, _REG_ESI);
  vm86->edi = READ_DWORD_REG (env, _REG_EDI);
  vm86->ebp = READ_DWORD_REG (env, _REG_EBP);
  vm86->esp = READ_DWORD_REG (env, _REG_ESP);
  vm86->cs = READ_SEGMENT_REG (env, _SEG_CS);
  vm86->ds = READ_SEGMENT_REG (env, _SEG_DS);
  vm86->es = READ_SEGMENT_REG (env, _SEG_ES);
  vm86->ss = READ_SEGMENT_REG (env, _SEG_SS);
  vm86->fs = READ_SEGMENT_REG (env, _SEG_FS);
  vm86->gs = READ_SEGMENT_REG (env, _SEG_GS);
  vm86->eip = env->eip;
  vm86->eflags = env->eflags;

  turtle_log = save;
  return exit;
}

void X86_stop ()
{
  if (current) {
    current->exit = 1;
  }
  else {
    linfo ("no virtual X86 context\n");
  }
}

void X86_dump_current ()
{
  if (current) {
    log ("virtual X86 context\n");
    log_emu (current);
  }
}

void X86_catch_VGA (int do_catch)
{
  catch_VGA = do_catch;
}

void X86_pre_init ()
{
  struct MUP_key keys[] =
    {{"turtle.log", TAG_STRING, {.sp = &turtle_log_name}}};

  MUP_add_keys (keys, LENOFA (keys));
}

void X86_init ()
{

  log_level = MUP_get_log_level ("x86");
  turtle_log_name = MUP_get_string ("turtle.log");

  if (turtle_log_name) {
    turtle_log = fopen (turtle_log_name, "w");
    if (NULL == turtle_log) {
      ERRFail ("failed to open `%s'", turtle_log_name);
    }

    setbuf (turtle_log, NULL);
  }
}

void cpu_register_physical_memory (unsigned long start_addr,
                                   unsigned long size,
                                   long phys_offset)
{
  DEREF (start_addr);
  DEREF (size);
  DEREF (phys_offset);
}

int cpu_register_io_memory(int io_index,
                           CPUReadMemoryFunc **mem_read,
                           CPUWriteMemoryFunc **mem_write)
{
  DEREF (io_index);

  vga_read_funcs = mem_read;
  vga_write_funcs = mem_write;

  turtle_write_vga_mem8 = mem_write[0];
  turtle_write_vga_mem16 = mem_write[1];
  turtle_write_vga_mem32 = mem_write[2];

  turtle_read_vga_mem8 = mem_read[0];
  turtle_read_vga_mem16 = mem_read[1];
  turtle_read_vga_mem32 = mem_read[2];
  return 0;
}
